
from typing import List
from typing import NewType
from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from copy import copy

from ogl.OglInheritance import OglInheritance
from ogl.OglInterface import OglInterface
from ogl.OglLink import OglLink
from ogl.OglObject import OglObject

from pyutmodel.PyutLinkType import PyutLinkType

from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.preferences.PluginPreferences import PluginPreferences

from pyutplugins.toolplugins.sugiyama.SugiyamaNode import SugiyamaNode
from pyutplugins.toolplugins.sugiyama.RealSugiyamaNode import RealSugiyamaNode
from pyutplugins.toolplugins.sugiyama.SugiyamaConstants import H_SPACE
from pyutplugins.toolplugins.sugiyama.SugiyamaConstants import LEFT_MARGIN
from pyutplugins.toolplugins.sugiyama.SugiyamaConstants import UP_MARGIN
from pyutplugins.toolplugins.sugiyama.SugiyamaConstants import V_SPACE
from pyutplugins.toolplugins.sugiyama.SugiyamaGlobals import SugiyamaGlobals
from pyutplugins.toolplugins.sugiyama.SugiyamaLink import SugiyamaLink
from pyutplugins.toolplugins.sugiyama.VirtualSugiyamaNode import VirtualSugiyamaNode

Nodes    = Union[RealSugiyamaNode, VirtualSugiyamaNode]
NodeList = NewType('NodeList', List[Nodes])
Levels   = NewType('Levels', List[NodeList])

HierarchicalGraphNode  = Union[RealSugiyamaNode, VirtualSugiyamaNode]
HierarchicalGraphNodes = NewType('HierarchicalGraphNodes', List[HierarchicalGraphNode])


class Sugiyama:
    """
    Automatic layout algorithm based on Sugiyama levels.

    This algorithm will change the class and links positions (not the
    structure). This plugin gives good result with diagrams that contain
    a lot of hierarchical relations (inheritance and interface), and poor
    association relations.

    """

    def __init__(self, pluginAdapter: IPluginAdapter):

        self.logger: Logger = getLogger(__name__)

        self._pluginAdapter: IPluginAdapter    = pluginAdapter
        self._preferences:   PluginPreferences = PluginPreferences()
        # Sugiyama nodes and links
        self.__realSugiyamaNodesList: List[RealSugiyamaNode] = []   # List of all RealSugiyamaNode's
        self._sugiyamaLinksList:      List[SugiyamaLink]     = []   # List of all SugiyamaLink's

        #  Hierarchy graph
        #  List of Real and Virtual Sugiyama nodes that take part in hierarchy
        # self._hierarchyGraphNodesList:    List[Union[RealSugiyamaNode, VirtualSugiyamaNode]] = []
        self._hierarchyGraphNodesList: HierarchicalGraphNodes = HierarchicalGraphNodes([])
        #  List of Sugiyama nodes that are not in a hierarchy
        self._nonHierarchyGraphNodesList: List[VirtualSugiyamaNode] = []
        self._nonHierarchyGraphLinksList: List[SugiyamaLink]        = []

        #  All nodes of the hierarchy are assigned to a level.
        #  A level is a list of nodes (real or virtual).
        # nodeList: NodeList = NodeList([])
        self._levels: Levels = Levels([])   # List of levels

    @property
    def levels(self) -> Levels:
        """
        For testability and security you only get a copy

        Returns:  Internal levels structure
        """
        levels: Levels = copy(self._levels)
        return levels

    def createInterfaceOglALayout(self, oglObjects):
        """
        Create the interface between oglObjects and Automatic Layout
        structure. A RealSugiyamaNode is created for each class, and a
        SugiyamaLink is created for each relation in the UML diagram.

        Args:
            oglObjects:  The Ogl Objects in the diagram
        """
        # Dictionary for oglObjects fast research
        # Key = OglObject, Value = RealSugiyamaNode
        dictOgl     = {}                # type: ignore
        # Dictionary for RealSugiyamaNode that takes part in hierarchy
        # Key = OglObject, Value = None
        dictSugiyamaHierarchy = {}      # type: ignore

        def createSugiyamaNode(theOglObject, theDictOgl):
            """
            Internal function for creating a RealSugiyamaNode and add it to
            self.__realSugiyamaNodesList and to dictOgl
            Args:
                theOglObject:
                theDictOgl:
            """
            # Create RealSugiyamaNode only if not already done
            if theOglObject not in theDictOgl:
                node = RealSugiyamaNode(theOglObject)
                self.__realSugiyamaNodesList.append(node)
                theDictOgl[theOglObject] = node

        def addNode2HierarchyGraph(theSugiyamaNode, theDictSugiyamaHierarchy):
            """
            Internal function for adding nodes that take part in hierarchy in
            the __hierarchyGraphNodesList.

            Args:
                theSugiyamaNode:
                theDictSugiyamaHierarchy:
            """
            if theSugiyamaNode not in theDictSugiyamaHierarchy:
                theDictSugiyamaHierarchy[theSugiyamaNode] = None
                self._hierarchyGraphNodesList.append(theSugiyamaNode)

        # For each OglObject or OglLink, create a specific interface
        for oglObject in oglObjects:

            # Class or Note :
            if isinstance(oglObject, OglObject):
                createSugiyamaNode(oglObject, dictOgl)      # type: ignore
            # Links
            elif isinstance(oglObject, OglLink):

                # Get source and destination oglObject
                srcOglClass = oglObject.sourceShape
                dstOglClass = oglObject.destinationShape

                # If the classes have not a RealSugiyamaNode attributed yet
                createSugiyamaNode(srcOglClass, dictOgl)
                createSugiyamaNode(dstOglClass, dictOgl)

                # Fix relations between nodes
                link = SugiyamaLink(oglObject)
                self._sugiyamaLinksList.append(link)
                srcSugiyamaNode = dictOgl[srcOglClass]
                dstSugiyamaNode = dictOgl[dstOglClass]
                link.setSource(srcSugiyamaNode)
                link.setDestination(dstSugiyamaNode)

                # If hierarchical link
                if isinstance(oglObject, OglInheritance) or isinstance(oglObject, OglInterface):

                    srcSugiyamaNode.addParent(dstSugiyamaNode, link)
                    dstSugiyamaNode.addChild(srcSugiyamaNode, link)

                    # Add nodes in list of hierarchical nodes
                    addNode2HierarchyGraph(srcSugiyamaNode, dictSugiyamaHierarchy)
                    addNode2HierarchyGraph(dstSugiyamaNode, dictSugiyamaHierarchy)

                # Non-hierarchical links
                else:

                    # Add link between source and destination interface
                    srcSugiyamaNode.addNonHierarchicalLink(dstSugiyamaNode, link)
                    dstSugiyamaNode.addNonHierarchicalLink(srcSugiyamaNode, link)

                    # Add link into non-hierarchical links' list
                    self._nonHierarchyGraphLinksList.append(link)

        # Create list of non-hierarchical nodes

        # For each class or note
        for sugiyamaNode in list(dictOgl.values()):
            # If not in hierarchy
            if sugiyamaNode not in dictSugiyamaHierarchy:
                self._nonHierarchyGraphNodesList.append(sugiyamaNode)

    def levelFind(self) -> bool:
        """
        Compute the best hierarchical level for each node.

        Returns: `True` if we computed the best, else `False` if there is a
        cycle in hierarchical links
        """
        nodesList: HierarchicalGraphNodes = self._hierarchyGraphNodesList
        # Fix nodes indexes corresponding to matrix column and line index
        nbNodes = len(nodesList)  # Number of nodes in hierarchy
        for i in range(nbNodes):
            nodesList[i].setIndex(i)
        # Initialize the boolean matrix
        #
        # Example of use:
        #
        #      |A|B|C
        #     -+-+-+-
        #     A|0|0|1 <-- that 1 means A is C's father
        #     -+-+-+-     and is on coordinates matrix[2][0]
        #     B|0|0|0
        #     -+-+-+-     matrix[column][line]
        #     C|0|1|0
        #
        #
        # noinspection PyUnusedLocal
        matrix = [[0 for el in range(nbNodes)] for el2 in range(nbNodes)]
        # Fill matrix
        # For each node in graph
        for node in nodesList:
            # For each father
            for (father, link) in node.getParents():
                # Mark relation with a '1' on coordinates[index Son][index Father]
                matrix[node.getIndex()][father.getIndex()] = 1      # type: ignore
        # Define levels

        # Sum each column of the matrix
        # Old code would not pass mypy or pycharm_projects
        # sumColumns = [None for el in range(nbNodes)]
        # el not used
        # noinspection PyUnusedLocal
        sumColumns: List[int] = [0 for el in range(nbNodes)]
        for i in range(nbNodes):
            sumColumns[i] = 0
            for el in matrix[i]:
                sumColumns[i] += el

        # Index of nodes that are not in any level yet
        indexNodes = list(range(nbNodes))
        # While not all nodes have an attributed level
        # while indexNodes != []:
        while indexNodes:
            # level = []  # Current level
            level: NodeList = NodeList([])
            indexNodesNotSel = indexNodes[:]
            indexNodesSel = []

            # For all nodes that haven't an attributed level
            for i in indexNodes:
                # When the sum of his matrix column is 0, that means he has no
                # parent or his parents are already in a level
                if sumColumns[i] == 0:
                    # The node is attributed on the current level
                    # Update the lists of selected nodes
                    indexNodesSel.append(i)
                    indexNodesNotSel.remove(i)
            # If no nodes is selected, there is a cycle in hierarchical links
            # if indexNodesSel == []:
            if not indexNodesSel:
                return False

            # For all the current level's nodes
            for i in indexNodesSel:
                level.append(nodesList[i])
                # Update the sum of the columns when we remove the line of the
                # node in the matrix
                for j in indexNodesNotSel:
                    sumColumns[j] -= matrix[j][i]
            # Update the list of the nodes that haven't a level yet
            indexNodes = indexNodesNotSel[:]
            # Add the current level to the list
            self._levels.append(level)

        # Fix nodes index and level for each node
        for idx in range(len(self._levels)):
            level = self._levels[idx]
            for i in range(len(level)):
                node = level[i]
                node.setIndex(i)
                node.setLevel(idx)

        # No error
        return True

    def addVirtualNodes(self):
        """
        Add a virtual node by level crossed between fathers and sons that are
        separated by more than one level.
        """
        # Internal function for updating a sons or fathers list
        def updateLink(nodesList, zLink, newNode):
            """
            Find the tuple (node, link2) in nodesList where link == link2 and
            replace node by newNode.
            """
            for i in range(len(nodesList)):
                (node, link2) = nodesList[i]
                if zLink == link2:
                    nodesList[i] = (newNode, zLink)
                    break

        # Add virtual nodes between a father and one of his sons
        def addVirtualNodesOnHierarchicalLink(zLink):

            srcNode = zLink.getSource()
            dstNode = zLink.getDestination()
            dstNodeLevel = dstNode.getLevel()

            # List of level index between dstNode and srcNode
            indexLevels = list(range(dstNodeLevel + 1, srcNode.getLevel()))

            # Continue only if there is at least one level between the two
            # nodes
            if len(indexLevels) == 0:
                return
            # noinspection PyUnusedLocal
            # For each crossed level, add a virtual node
            virtualNodes = [VirtualSugiyamaNode() for el in indexLevels]

            # Fix level
            for i in range(len(virtualNodes)):
                virtualNode: VirtualSugiyamaNode = virtualNodes[i]
                virtualNode.setLevel(dstNodeLevel + i + 1)

            # Fix relation between virtual nodes
            for i in range(len(virtualNodes) - 1):
                virtualNodes[i].addChild(virtualNodes[i + 1], zLink)
                virtualNodes[i + 1].addParent(virtualNodes[i], zLink)

            # Fix relations between virtual and real nodes
            virtualNodes[-1].addChild(srcNode, zLink)
            virtualNodes[0].addParent(dstNode, zLink)

            updateLink(dstNode.getChildren(), zLink, virtualNodes[0])
            updateLink(srcNode.getParents(), zLink, virtualNodes[-1])

            # Add virtual nodes in levels
            for i in range(len(virtualNodes)):
                level = self._levels[dstNodeLevel + i + 1]
                level.append(virtualNodes[i])
                # Fix index of the virtual node
                level[-1].setIndex(len(level) - 1)

            # Add virtual nodes in link in order bottom to top
            for i in range(len(virtualNodes) - 1, -1, -1):
                zLink.addVirtualNode(virtualNodes[i])

        # For all links
        for link in self._sugiyamaLinksList:
            # If hierarchical link
            if link.getType() == PyutLinkType.INHERITANCE or link.getType() == PyutLinkType.INTERFACE:

                # Add virtual nodes
                addVirtualNodesOnHierarchicalLink(link)

    def barycenter(self):
        """
        Find nodes index for minimizing hierarchical links crossing.
        """
        MAX_ITER = 20

        while self._getNbIntersectAll() > 0 and MAX_ITER > 0:

            # Downward phase

            # For each level except first
            for i in range(1, len(self._levels)):

                # Compute parents down-barycenter
                if i > 0:
                    self._downBarycenterLevel(i - 1)
                # Compute sons up-barycenter
                if i < len(self._levels) - 1:
                    self._upBarycenterLevel(i + 1)

                # Compute up-barycenter on current level
                self._upBarycenterLevel(i)
                self._sortLevel(i)
                self._shiftSameBarycenter(i)

            # Upward phase

            if self._getNbIntersectAll() > 0:

                indexList = list(range(len(self._levels) - 1))
                indexList.reverse()
                for i in indexList:

                    self._downBarycenterLevel(i)
                    self._sortLevel(i)
                    self._shiftSameBarycenter(i)

                MAX_ITER -= 1

    def addNonHierarchicalNodes(self):
        """
        Add non-hierarchical nodes into levels.
        """
        # Vocabulary:
        # Internal node: nodes present in levels
        # External node: nodes not present in levels yet

        # Dictionary internalNodes externalNodes:
        # Keys are internal respectively external nodes
        # Values :
        #   - externalNodes : # of zLink to internal nodes
        internalNodes = {}  # type: ignore
        externalNodes = {}

        # Make dictionary of internal nodes
        for node in self._hierarchyGraphNodesList:
            internalNodes[node] = None

        # Make dictionary of external nodes
        for node in self._nonHierarchyGraphNodesList:
            # Count zLink to internal nodes
            count = 0
            for (dstNode, link) in node.getNonHierarchicalLink():
                if dstNode in internalNodes:
                    count += 1

            # Add node to externalNodes
            externalNodes[node] = count

        # If there is no level (no inheritance or realisation) but there are
        # nodes to put in, create new level
        if not self._levels and externalNodes:
            # Add one level for nodes
            self._levels.append(NodeList([]))

        # Function for getting node that has most connections to internal
        # nodes
        def mostConnection(zExternalNodes):

            maxNode    = None
            maxNbLinks = -1
            for (nbLinkNode, nbLinks) in list(zExternalNodes.items()):
                # If current node has more connections
                if nbLinks > maxNbLinks:
                    maxNode    = nbLinkNode
                    maxNbLinks = nbLinks

            return maxNode
        # Function for evaluating best level and best index for an external
        # node

        def bestPos(zExtNode, zInternalNodes):
            """
            Evaluate average of level of linked nodes
            Args:
                zExtNode:
                zInternalNodes:

            Returns: (level, index)
            """
            nb:        int = 0
            summation: int = 0
            nodes     = []  # List of connected internal nodes

            # For all non-hierarchical links
            for (zDstNode, zLink) in zExtNode.getNonHierarchicalLink():
                # If node linked to internal nodes
                if zDstNode in zInternalNodes:
                    # Add connected node to list
                    nodes.append(zDstNode)
                    # Add level to sum and count number of zLink
                    summation += zDstNode.getLevel()
                    nb += 1
            # If no zLink to internal nodes
            # if nodes == []:
            if not nodes:
                return None, None

            # Find closer node to average position
            avgLevel = summation // nb
            levelNodes = []  # List of nodes on same level
            bestLevel = None
            # Fix best level on first node
            if nodes:
                bestLevel = nodes[0].getLevel()

            # For all connected internal nodes
            for connectedInternalNode in nodes:
                nodeLevel = connectedInternalNode.getLevel()

                # If current node is on bestLevel
                if nodeLevel == bestLevel:
                    levelNodes.append(connectedInternalNode)

                # Else if current node is nearer to average position or
                # is at same distance but with less nodes on level
                # TODO Refactor this test to a method
                elif abs(nodeLevel - avgLevel) < abs(bestLevel - avgLevel) or (abs(nodeLevel - avgLevel) == abs(bestLevel - avgLevel) and       # type: ignore
                                                                               len(self._levels[nodeLevel]) < len(self._levels[bestLevel])):    # type: ignore

                    # Store best level
                    bestLevel = nodeLevel

                    # Start new list of nodes on new best level
                    levelNodes = [connectedInternalNode]

            # Return average of nodes' level
            return bestLevel, levelNodes[len(levelNodes) // 2].getIndex()

        # Function for getting level that has fewer nodes.
        def getLessFilledLevel():

            lessLevel = 0  # Index of level that has less node in it
            nb = len(self._levels[lessLevel])

            for x in range(1, len(self._levels)):
                if len(self._levels[x]) < nb:
                    lessLevel = x
                    nb = len(self._levels[x])

            return lessLevel

        # Function to move a node from internal to external nodes.
        def moveExternal2Internal(zNode, zInternalNodes, zExternalNodes):

            # Remove node from external nodes
            del zExternalNodes[zNode]

            # Add node to internal nodes
            zInternalNodes[zNode] = None

            # For all his linked external nodes, update their counter
            # zExtNode = None   NOT USED
            for (zDstNode, zLink) in zNode.getNonHierarchicalLink():
                if zDstNode in zExternalNodes:
                    zExternalNodes[zDstNode] += 1

        # While there are nodes still not in hierarchy
        while externalNodes:
            # Get external node that has most connections to internalNodes
            extNode = mostConnection(externalNodes)

            self.logger.info(f'extNode.getName(): `{extNode.getName()}`')
            # Evaluate best level and index for the node
            (level, index) = bestPos(extNode, internalNodes)

            # If node has no connection to internal node
            if level is None:
                # Find level that is less filled of nodes
                level = getLessFilledLevel()
                index = len(self._levels[level])

            # Add node in levels
            extNode.setLevel(level)
            extNode.setIndex(index)
            self._levels[level].insert(index, extNode)
            # Shift index attributes on right
            for i in range(index + 1, len(self._levels[level])):
                self._levels[level][i].setIndex(i)

            # Move node from external to internal nodes
            moveExternal2Internal(extNode, internalNodes, externalNodes)

    def fixPositions(self):
        """
        Compute coordinates for nodes and links.
        """
        self._fixNodesNeighbors()
        self._fixNodesPositions()
        self._fixLinksPositions()

    def _downBarycenterLevel(self, indexLevel):
        """
        Compute down barycenter (from sons) for all nodes on level.

        Args:
            indexLevel:   The index level
        """
        level = self._levels[indexLevel]
        for node in level:
            node.downBarycenterIndex()

    def _upBarycenterLevel(self, indexLevel):
        """
        Compute up barycenter (from parents) for all nodes on level.

        @param indexLevel : index of level
        @author Nicolas Dubois
        """
        level = self._levels[indexLevel]
        for node in level:
            node.upBarycenterIndex()

    def _shiftSameBarycenter(self, indexLevel):
        """
        Do a left circular shifting on nodes with same barycenter on a level.

        For each group of nodes which have the same pre-calculated value, do a
        left circular shifting of the nodes.

        @param indexLevel : index of level
        @author Nicolas Dubois
        """
        level = self._levels[indexLevel]

        # Save current level
        levelSaved = level[:]
        # Count crossings
        nbIntersections = self._getNbIntersectAll()

        # Shift same barycenter
        for i in range(len(level) - 1):
            if level[i].getBarycenter() is not None and level[i].getBarycenter() == level[i + 1].getBarycenter():

                level.insert(i, level.pop(i + 1))
                # Fix index
                level[i].setIndex(i)
                level[i + 1].setIndex(i + 1)

        # If new order give more intersections, return to old order
        if self._getNbIntersectAll() > nbIntersections:
            self._levels[indexLevel] = levelSaved
            for i in range(len(levelSaved)):
                levelSaved[i].setIndex(i)

    def _sortLevel(self, indexLevel):
        """
        Sort nodes on a level according to pre-calculated barycenter value.
        Nodes that don't have a barycenter value keep their place.

        Args:
            indexLevel:  index of level in self.__levels to sort
        """
        level = self._levels[indexLevel]
        levelCopy = level[:]

        nbIntersect = self._getNbIntersectAll()

        # Get list of nodes who have a barycenter value
        listIndex = []
        for i in range(len(levelCopy)):
            if levelCopy[i].getBarycenter() is not None:
                listIndex.append(i)

        # Create list of nodes to sort
        listNodes = []
        for i in listIndex:
            listNodes.append(levelCopy[i])

        # Sort list of nodes
        listNodes.sort(key=SugiyamaGlobals.cmpBarycenter)

        # Put sorted list in levelCopy
        for i in range(len(listNodes)):
            levelCopy[listIndex[i]] = listNodes[i]

        # Fix indexes
        for i in range(len(levelCopy)):
            levelCopy[i].setIndex(i)

        # If there are more intersections than before, keep original order
        nbIntersect2 = self._getNbIntersectAll()
        if nbIntersect < nbIntersect2:
            # Fix indexes
            for i in range(len(level)):
                level[i].setIndex(i)
        else:
            # Else set new order
            self._levels[indexLevel] = levelCopy
        # nbIntersect3 = self.__getNbIntersectAll()    NOT USED

    def _getNbIntersectAll(self):
        """

        Returns:  The number of intersections between hierarchy relations.
        """
        count = 0
        for i in range(len(self._levels) - 1):
            count += self._getNbIntersect2Levels(i)

        return count

    def _getNbIntersect2Levels(self, upperLevel):
        """
        The two levels index are [upperLevel] and [upperLevel + 1].

        Args:
            upperLevel:  index of upper level

        Returns: The intersections number of hierarchical links between two levels.
        """
        # Get nodes from the level
        nodes = self._levels[upperLevel]

        # Count intersect
        count = 0

        # For each node of the layer
        for indFatherL in range(len(nodes) - 1):
            # For each son of the current node
            for (sonL, link) in nodes[indFatherL].getChildren():

                # Check intersect with all next parents
                indFatherR = indFatherL + 1
                while indFatherR < len(nodes):
                    for (sonR, rLink) in nodes[indFatherR].getChildren():
                        # If intersect
                        if sonL.getIndex() > sonR.getIndex():
                            count += 1
                    indFatherR += 1

        return count

    def _fixNodesNeighbors(self):
        """
        For each node, fix his right neighbor.

        @author Nicolas Dubois
        """
        # For each node, fix his neighbors if he has, None else
        for level in self._levels:
            nbNodes = len(level)

            for i in range(nbNodes - 1):
                level[i + 1].setLeftNode(level[i])
                level[i].setRightNode(level[i + 1])

            # For first and last nodes of the level
            level[0].setLeftNode(None)      # type: ignore
            level[nbNodes - 1].setRightNode(cast(SugiyamaNode, None))

    def _fixNodesPositions(self):

        # Compute start positions packed on left
        y: int = UP_MARGIN
        for level in self._levels:
            x:         int = LEFT_MARGIN
            maxHeight: int = 0

            for node in level:
                (width, height) = node.getSize()
                node.setPosition(x, y)
                x += width + H_SPACE
                maxHeight = max(maxHeight, height)
            y += maxHeight + V_SPACE

        if self._preferences.sugiyamaStepByStep is True:
            SugiyamaGlobals.waitKey(self._pluginAdapter, optionalMessage=None)
        else:
            self.logger.info(f'.__fixNodesPositions() is complete')

        # While nodes have to be moved
        moved: bool = True
        while moved:
            moved = False
            # Compute average coordinates for each node
            for level in self._levels:
                for node in level:
                    if node.balance():
                        moved = True
                        msg: str = f'LEVEL - node: {node} {level}'
                        if self._preferences.sugiyamaStepByStep is True:
                            SugiyamaGlobals.waitKey(self._pluginAdapter, msg)
                        else:
                            self.logger.info(msg)

    def _fixLinksPositions(self):
        """
        Compute links new positions.
        """
        # For each hierarchical link, fix anchors coordinates
        for level in self._levels:
            for node in level:
                node.fixAnchorPos()

        # For each hierarchical link, add control points to pass through
        # each virtual node
        for link in self._sugiyamaLinksList:
            link.fixControlPoints()




from pyutmodel.PyutLinkType import PyutLinkType


class ALayoutLink:
    """
    ALayoutLink : Interface between pyut/ogl link and ALayout algorithms.

    ALayout algorithms can use this interface to access the links of the
    diagram. The first reason is that the interface protects the structure
    of the diagram. The second is that pyut structure and methods could
    be changed. In a such case, the only files to update is the interface, not
    your automatic layout algorithm.

    """
    def __init__(self, oglLink):
        """
        Constructor.

        @author Nicolas Dubois
        """
        self._oglLink = oglLink
        self.__srcNode = None
        self.__dstNode = None

    def setSource(self, node):
        """
        Set the source node.

        @param InterfaceSugiyamaNode node: source node of the link
        @author Nicolas Dubois
        """
        self.__srcNode = node

    def getSource(self):
        """
        Return the source node.

        @return InterfaceSugiyamaNode: source node of the link
        @author Nicolas Dubois
        """
        return self.__srcNode

    def setDestination(self, node):
        """
        Set the destination node.

        @param InterfaceSugiyamaNode node: destination node of the link
        @author Nicolas Dubois
        """
        self.__dstNode = node

    def getDestination(self):
        """
        Return the destination node.

        @return InterfaceSugiyamaNode: destination node of the link
        @author Nicolas Dubois
        """
        return self.__dstNode

    def setSrcAnchorPos(self, x: int, y: int):
        """
        Set anchor position (absolute coordinates) on source class.

        Args:
            x:
            y:
        """
        self._oglLink.sourceAnchor.SetPosition(x, y)

    def getSrcAnchorPos(self):
        """
        Get anchor position (absolute coordinates) on source class.

        Returns:    (int, int) : tuple with (x, y) coordinates
        """

        return self._oglLink.sourceAnchor.GetPosition()

    def setDestAnchorPos(self, x: int, y: int):
        """
        Set anchor position (absolute coordinates) on destination class.

        Args:
            x:
            y:
        """
        self._oglLink.destinationAnchor.SetPosition(x, y)

    def getDestAnchorPos(self):
        """
        Return anchor position (absolute coordinates) on destination class.

        @return (float, float) : tuple with (x, y) coordinates
        @author Nicolas Dubois
        """
        return self._oglLink.GetDestination().GetPosition()

    def addControlPoint(self, control, last=None):
        """
        Add a control point. If param last given, add point right after last.

        @param ControlPoint control : control point to add
        @param ControlPoint last    : add control right after last
        @author Nicolas Dubois
        """
        self._oglLink.AddControl(control, last)

    def removeControlPoint(self, controlPoint):
        """
        Remove a control point.

        @param ControlPoint controlPoint: control point to remove
        @author Nicolas Dubois
        """
        self._oglLink.Remove(controlPoint)

    def removeAllControlPoints(self):
        """
        Remove all control points.

        @author Nicolas Dubois
        """
        self._oglLink.RemoveAllControlPoints()

    def getType(self) -> PyutLinkType:
        """
        Return the link type

        Returns: Link type
        """
        return self._oglLink.pyutObject.linkType




from typing import Dict
from typing import List
from typing import NewType

from enum import Enum

from dataclasses import dataclass
from dataclasses import field

#
#  Both of these hold the class plugintypes for the Plugins
#
PluginList  = NewType('PluginList',  List[type])
PluginIDMap = NewType('PluginIDMap', Dict[int, type])


def createPlugIdMapFactory() -> PluginIDMap:
    return PluginIDMap({})


PluginName        = NewType('PluginName', str)
FormatName        = NewType('FormatName', str)
PluginExtension   = NewType('PluginExtension', str)
PluginDescription = NewType('PluginDescription', str)


class PluginMapType(Enum):
    INPUT_MAP  = 'InputMap'
    OUTPUT_MAP = 'OutputMap'
    TOOL_MAP   = 'ToolMap'


#
# Some nice syntactic sugar
#
@dataclass
class BasePluginMap:
    mapType:     PluginMapType = PluginMapType.INPUT_MAP
    pluginIdMap: PluginIDMap   = field(default_factory=createPlugIdMapFactory)

    def __init__(self):
        self.pluginIdMap = PluginIDMap({})

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'{self.mapType} plugin count: {len(self.pluginIdMap)}'


@dataclass
class InputPluginMap(BasePluginMap):
    def __init__(self):
        super().__init__()
        self.mapType = PluginMapType.INPUT_MAP

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'{self.mapType} plugin count: {len(self.pluginIdMap)}'


@dataclass
class OutputPluginMap(BasePluginMap):
    def __init__(self):
        super().__init__()
        self.mapType = PluginMapType.OUTPUT_MAP


@dataclass
class ToolsPluginMap(BasePluginMap):
    def __init__(self):
        super().__init__()
        self.mapType = PluginMapType.TOOL_MAP




from typing import Set
from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from miniogl.AnchorPoint import AnchorPoint
from miniogl.LinePoint import LinePoint
from miniogl.LineShape import ControlPoints

from pyutmodel.PyutObject import PyutObject

from ogl.OglObject import OglObject
from ogl.OglClass import OglClass
from ogl.OglLink import OglLink
from ogl.OglNote import OglNote

from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.ioplugins.gml.UnsupportedOperation import UnsupportedOperation


class GMLExporter:

    VERSION: str = '1.0.0'

    GRAPH_TOKEN: str = 'graph'

    ID_TOKEN:    str = 'id'
    LABEL_TOKEN: str = 'label'
    NODE_TOKEN:  str = 'node'
    EDGE_TOKEN:  str = 'edge'

    SOURCE_ID_TOKEN: str = 'source'
    TARGET_ID_TOKEN: str = 'target'

    GRAPHICS_TOKEN: str = 'graphics'
    START_TOKEN:    str = '['
    END_TOKEN:      str = ']'

    QUOTE_TOKEN: str = '"'

    LINE_DEFINITION_TOKEN:  str = 'Line'
    POINT_DEFINITION_TOKEN: str = 'point'

    X_POSITION_TOKEN: str = 'x'
    Y_POSITION_TOKEN: str = 'y'
    Z_POSITION_TOKEN: str = 'z'
    WIDTH_TOKEN:      str = 'w'
    HEIGHT_TOKEN:     str = 'h'
    DEPTH_TOKEN:      str = 'd'

    singleTab: str = ''
    doubleTab: str = ''
    tripleTab: str = ''
    quadrupleTab: str = ''
    quintupleTab: str = ''

    def __init__(self):

        self.logger:   Logger = getLogger(__name__)
        self._gml:     str    = ''

        self._prettyPrint: bool = True

    def translate(self, umlObjects: OglObjects):

        if self._prettyPrint is True:
            GMLExporter.singleTab = '\t'
            GMLExporter.doubleTab = '\t\t'
            GMLExporter.tripleTab = '\t\t\t'
            GMLExporter.quadrupleTab = '\t\t\t\t'
            GMLExporter.quintupleTab = '\t\t\t\t\t'

        gml: str = self._generateGraphStart()
        gml = self._generateNodes(umlObjects, gml)
        gml = self._generateUniqueEdges(umlObjects=umlObjects, gml=gml)

        gml = self._generateGraphTermination(gml)
        self._gml = gml

    @property
    def gml(self):
        return self._gml

    @gml.setter
    def gml(self, theNewValue):
        raise UnsupportedOperation('gml is a read-only property')

    @property
    def prettyPrint(self):
        return self._prettyPrint

    @prettyPrint.setter
    def prettyPrint(self, theNewValue):
        self._prettyPrint = theNewValue

    def write(self, pathToFile: str):

        with open(pathToFile, 'w') as writer:
            writer.write(self._gml)

    def _generateNodes(self, umlObjects: OglObjects, gml: str) -> str:

        nodeGml: str = ''
        for umlClass in umlObjects:
            if isinstance(umlClass, OglClass) or isinstance(umlClass, OglNote):
                oglObject:  OglObject  = cast(OglObject, umlClass)
                pyutObject: PyutObject = oglObject.pyutObject
                nodeGml = (
                    f'{nodeGml}'
                    f'{GMLExporter.singleTab}{GMLExporter.NODE_TOKEN} {GMLExporter.START_TOKEN}\n'
                    f'{GMLExporter.doubleTab}{GMLExporter.ID_TOKEN} {oglObject.id}\n'
                    f'{GMLExporter.doubleTab}{GMLExporter.LABEL_TOKEN} "{pyutObject.name}"\n'
                    f'{self._generateNodeGraphicsSection(oglObject)}'
                    f'{GMLExporter.singleTab}{GMLExporter.END_TOKEN}\n'
                )
                self.logger.debug(f'{nodeGml}')
        return f'{gml}{nodeGml}'

    def _generateNodeGraphicsSection(self, oglObject: OglObject) -> str:

        pos = oglObject.GetPosition()
        x = pos[0]
        y = pos[1]
        z = 0
        dimensions = oglObject.GetSize()
        w = dimensions[0]
        h = dimensions[1]
        d = 0
        gml = (
            f'{GMLExporter.doubleTab}{GMLExporter.GRAPHICS_TOKEN} {GMLExporter.START_TOKEN}\n'
            
            f'{GMLExporter.tripleTab}{GMLExporter.X_POSITION_TOKEN} {x}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.Y_POSITION_TOKEN} {y}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.Z_POSITION_TOKEN} {z}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.WIDTH_TOKEN} {w}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.HEIGHT_TOKEN} {h}\n'
            f'{GMLExporter.tripleTab}{GMLExporter.DEPTH_TOKEN} {d}\n'
            f'{GMLExporter.tripleTab}type "rectangle"\n'
            f'{GMLExporter.tripleTab}width 0.12\n'
            f'{GMLExporter.tripleTab}fill "#ff0000"\n'
            f'{GMLExporter.tripleTab}outline "#000000"\n'
            
            f'{GMLExporter.doubleTab}{GMLExporter.END_TOKEN}\n'
        )
        return gml

    def _generateGraphStart(self, graphName: str = 'DefaultGraphName') -> str:

        gml: str = (
            f'{GMLExporter.GRAPH_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.singleTab}directed 1\n'
            f'{GMLExporter.singleTab}version  1.0\n'
            f'{GMLExporter.singleTab}label "GML for {graphName}"\n'
            f'{GMLExporter.singleTab}comment "Generated by GML Exporter Version {GMLExporter.VERSION}"\n'
        )

        return gml

    def _generateGraphTermination(self, gml: str) -> str:

        gml = f'{gml}{GMLExporter.END_TOKEN}'
        return gml

    def _generateUniqueEdges(self, umlObjects: OglObjects, gml: str) -> str:

        linkSet:  Set    = set()        # Concatenated str link ids;  e.g, 1-2

        for umlClass in umlObjects:
            if isinstance(umlClass, OglClass) or isinstance(umlClass, OglNote):
                oglObject: OglObject = cast(OglObject, umlClass)
                links = oglObject.links
                self.logger.info(f'links: {links}')
                for oglLink in links:
                    srcOglId:  int = oglLink.sourceShape.id
                    destOglId: int = oglLink.destinationShape.id
                    linkIds:   str = f'{srcOglId}-{destOglId}'
                    if linkIds not in linkSet:
                        gml = self.__generateUniqueEdge(oglLink=oglLink, gml=gml)
                        linkSet.add(linkIds)

        return gml

    def __generateUniqueEdge(self, oglLink: OglLink, gml: str) -> str:

        srcOglId:  int = oglLink.sourceShape.id
        destOglId: int = oglLink.destinationShape.id

        gml = (
            f'{gml}'
            f'{GMLExporter.singleTab}{GMLExporter.EDGE_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.doubleTab}{GMLExporter.ID_TOKEN} {oglLink.id}\n'
            f'{GMLExporter.doubleTab}{GMLExporter.SOURCE_ID_TOKEN} {srcOglId}\n'
            f'{GMLExporter.doubleTab}{GMLExporter.TARGET_ID_TOKEN} {destOglId}\n'
            f'{self.__generateEdgeGraphicsSection(oglLink=oglLink)}'
            f'{GMLExporter.singleTab}{GMLExporter.END_TOKEN}\n'
        )

        return gml

    def __generateEdgeGraphicsSection(self, oglLink: OglLink) -> str:

        srcAnchor:  AnchorPoint = oglLink.sourceAnchor
        destAnchor: AnchorPoint = oglLink.destinationAnchor

        controlPoints: ControlPoints = oglLink.GetControlPoints()

        edgeGml: str = (
            f'{GMLExporter.doubleTab}{GMLExporter.GRAPHICS_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.tripleTab}type "line"\n'
            f'{GMLExporter.tripleTab}arrow "last"\n'
            f'{GMLExporter.tripleTab}{GMLExporter.LINE_DEFINITION_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{self.__generatePoint(srcAnchor)}'
            f'{self.__generatePoints(controlPoints)}'
            f'{self.__generatePoint(destAnchor)}'
            f'{GMLExporter.tripleTab}{GMLExporter.END_TOKEN}\n'
            f'{GMLExporter.doubleTab}{GMLExporter.END_TOKEN}\n'
        )

        return edgeGml

    def __generatePoints(self, points: ControlPoints) -> str:

        pointsGml: str = ''
        for point in points:
            pointsGml = (
                f'{pointsGml}{self.__generatePoint(point)}'
            )

        return pointsGml

    def __generatePoint(self, linePoint: LinePoint) -> str:

        position: Tuple[int, int] = linePoint.GetPosition()

        x:        int = position[0]
        y:        int = position[1]
        z:        int = 0
        pointGml: str = (
            f'{GMLExporter.quadrupleTab}{GMLExporter.POINT_DEFINITION_TOKEN} {GMLExporter.START_TOKEN}\n'
            f'{GMLExporter.quintupleTab}{GMLExporter.X_POSITION_TOKEN} {x}\n'
            f'{GMLExporter.quintupleTab}{GMLExporter.Y_POSITION_TOKEN} {y}\n'
            f'{GMLExporter.quintupleTab}{GMLExporter.Z_POSITION_TOKEN} {z}\n'
            f'{GMLExporter.quadrupleTab}{GMLExporter.END_TOKEN}\n'
        )

        return pointGml




from typing import Optional

from wx import DD_NEW_DIR_BUTTON
from wx import FD_OPEN
from wx import FD_MULTIPLE
from wx import FD_CHANGE_DIR
from wx import FD_FILE_MUST_EXIST
from wx import FD_OVERWRITE_PROMPT
from wx import FD_SAVE
from wx import ICON_ERROR
from wx import ID_CANCEL
from wx import OK

from wx import DirDialog
from wx import FileDialog
from wx import FileSelector
from wx import MessageDialog
from wx import Yield as wxYield

from pyutplugins.preferences.PluginPreferences import PluginPreferences

from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglLinks
from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat
from pyutplugins.plugintypes.ExportDirectoryResponse import ExportDirectoryResponse
from pyutplugins.plugintypes.ImportDirectoryResponse import ImportDirectoryResponse
from pyutplugins.plugintypes.MultipleFileRequestResponse import MultipleFileRequestResponse
from pyutplugins.plugintypes.SingleFileRequestResponse import SingleFileRequestResponse

from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.PluginDataTypes import PluginName

UNSPECIFIED_NAME:        FormatName        = FormatName('Unspecified Plugin Name')
UNSPECIFIED_EXTENSION:   PluginExtension   = PluginExtension('*')
UNSPECIFIED_DESCRIPTION: PluginDescription = PluginDescription('Unspecified Plugin Description')


class BasePluginInterface:
    """
    This is meant to provide base properties and methods for the Input/Output
    pyutplugins and the Tool Plugins

    Implementations set the protected variables at class construction

    There should be no implementations of this interface
    """

    def __init__(self, pluginAdapter: IPluginAdapter):
        """
        Menu handlers may instantiate a plugin merely to get plugin information.  In that case,
        the input parameter will be None

        Args:
            pluginAdapter:   A class that implements ICommunicator

        """
        self._pluginAdapter:     IPluginAdapter      = pluginAdapter
        self._pluginPreferences: PluginPreferences = PluginPreferences()

        #
        # To be set by implementor constructor and read by property
        self._name:         PluginName = PluginName('Implementor must provide the plugin name')
        self._author:       str = 'Implementor must provide the plugin author'
        self._version:      str = 'Implementor must provide the version'
        self._inputFormat:  InputFormat  = InputFormat(formatName=UNSPECIFIED_NAME, extension=UNSPECIFIED_EXTENSION, description=UNSPECIFIED_DESCRIPTION)
        self._outputFormat: OutputFormat = OutputFormat(formatName=UNSPECIFIED_NAME, extension=UNSPECIFIED_EXTENSION, description=UNSPECIFIED_DESCRIPTION)

    @property
    def name(self) -> PluginName:
        """
        Implementations set the protected variable at class construction

        Returns:  The plugin name
        """
        return self._name

    @property
    def author(self) -> str:
        """
        Implementations set the protected variable at class construction

        Returns:  The author's name
        """
        return self._author

    @property
    def version(self) -> str:
        """
        Implementations set the protected variable at class construction

        Returns: The plugin version string
        """
        return self._version

    @property
    def inputFormat(self) -> InputFormat:
        """
        Implementations set the protected variable at class construction

        Returns: The input format type; Plugins should return `None` if they do
        not support input operations
        """
        return self._inputFormat

    @property
    def outputFormat(self) -> OutputFormat:
        """
        Implementations set the protected variable at class construction

        Returns: The output format type;  Plugins should return `None` if they do
        not support output operations
        """
        return self._outputFormat

    @classmethod
    def displayNoUmlFrame(cls):
        booBoo: MessageDialog = MessageDialog(parent=None, message='No UML frame', caption='Try Again!', style=OK | ICON_ERROR)
        booBoo.ShowModal()

    @classmethod
    def displayNoSelectedOglObjects(cls):
        booBoo: MessageDialog = MessageDialog(parent=None, message='No selected UML objects', caption='Try Again!', style=OK | ICON_ERROR)
        booBoo.ShowModal()

    def askForFileToImport(self, startDirectory: str | None) -> SingleFileRequestResponse:
        """
        Called by plugin to ask for a file to import

        Args:
            startDirectory: The directory to display

        Returns:  The request response
        """
        defaultDir:  Optional[str] = startDirectory

        if defaultDir is None:
            defaultDir = self._pluginAdapter.currentDirectory
        file = FileSelector(
            "Choose a file to import",
            # wildcard=inputFormat.name + " (*." + inputFormat.extension + ")|*." + inputFormat.description,
            default_path=defaultDir,
            wildcard=self.__composeWildCardSpecification(),
            flags=FD_OPEN | FD_FILE_MUST_EXIST | FD_CHANGE_DIR
        )
        response: SingleFileRequestResponse = SingleFileRequestResponse()
        if file == '':
            response.cancelled = True
            response.fileName  = ''
        else:
            response.cancelled = False
            response.fileName = file

        return response

    def askToImportMultipleFiles(self, startDirectory: str | None) -> MultipleFileRequestResponse:
        """
        This method determines how to filter the input files via the Plugin Input
        format specification.
        Args:
            startDirectory:   The initial directory to display

        Returns:  The request response
        """
        defaultDir:  Optional[str] = startDirectory

        if defaultDir is None:
            defaultDir = self._pluginAdapter.currentDirectory

        dlg: FileDialog = FileDialog(
            None,
            "Choose files to import",
            wildcard=self.__composeWildCardSpecification(),
            defaultDir=defaultDir,
            style=FD_OPEN | FD_FILE_MUST_EXIST | FD_MULTIPLE | FD_CHANGE_DIR
        )
        dlg.ShowModal()
        response: MultipleFileRequestResponse = MultipleFileRequestResponse()
        if dlg.GetReturnCode() == ID_CANCEL:
            response.directoryName = ''
            response.fileList      = []
            response.cancelled     = True
        else:
            response.directoryName = dlg.GetDirectory()
            response.fileList      = dlg.GetFilenames()
            response.cancelled     = False

        return response

    def askForFileToExport(self, defaultFileName: str = '') -> SingleFileRequestResponse:
        """
        Called by a plugin to ask for the export file name

        Returns: The appropriate response object
        """
        wxYield()

        outputFormat: OutputFormat = self.outputFormat

        wildCard:    str = f'{outputFormat.formatName} (*.{outputFormat.extension}) |*.{outputFormat.extension}'
        fileName:    str = FileSelector("Choose export file name",
                                        default_filename=defaultFileName,
                                        wildcard=wildCard,
                                        flags=FD_SAVE | FD_OVERWRITE_PROMPT | FD_CHANGE_DIR)

        response: SingleFileRequestResponse = SingleFileRequestResponse(cancelled=False)
        if fileName == '':
            response.fileName  = ''
            response.cancelled = True
        else:
            response.fileName = fileName

        return response

    def askForImportDirectoryName(self) -> ImportDirectoryResponse:
        """
        Called by plugin to ask which directory must be imported

        Returns:  The appropriate response object;  The directory name is valid only if
        response.cancelled is True
        """
        dirDialog: DirDialog = DirDialog(None,
                                         "Choose a directory to import",
                                         defaultPath=self._pluginAdapter.currentDirectory,
                                         style=DD_NEW_DIR_BUTTON)

        response: ImportDirectoryResponse = ImportDirectoryResponse()
        if dirDialog.ShowModal() == ID_CANCEL:
            response.cancelled     = True
            response.directoryName = ''
        else:
            response.directoryName = dirDialog.GetPath()
            response.cancelled     = False
            # self._pluginAdapter.currentDirectory = response.directoryName    # TODO: Should plugin be doing this?  No

        dirDialog.Destroy()

        return response

    def askForExportDirectoryName(self, preferredDefaultPath: str | None) -> ExportDirectoryResponse:
        """
        Called by plugin to ask for an output directory
        Args:
            preferredDefaultPath:

        Returns:  The appropriate response object;  The directory name is valid only if
        response.cancelled is True
        """
        if preferredDefaultPath is None:
            defaultPath: str = self._pluginAdapter.currentDirectory
        else:
            defaultPath = preferredDefaultPath

        dirDialog: DirDialog = DirDialog(None, "Choose a destination directory", defaultPath=defaultPath)

        response: ExportDirectoryResponse = ExportDirectoryResponse(cancelled=False)
        if dirDialog.ShowModal() == ID_CANCEL:
            dirDialog.Destroy()
            response.directoryName = ''
            response.cancelled     = True
        else:
            directory = dirDialog.GetPath()
            # self._pluginAdapter.currentDirectory = directory     # TODO  Should a plugin do this;  No
            dirDialog.Destroy()
            response.directoryName = directory

        return response

    def _layoutUmlClasses(self, oglClasses: OglClasses):
        """
        Organize by vertical descending sizes

        Args:
            oglClasses
        """
        # Sort by descending height
        # noinspection PyProtectedMember
        sortedOglClasses = sorted(oglClasses, key=lambda oglClassToSort: oglClassToSort._height, reverse=True)

        x: int = 20
        y: int = 20

        incY: int = 0
        for oglClass in sortedOglClasses:
            incX, sy = oglClass.GetSize()
            incX += 20
            sy += 20
            incY = max(incY, int(sy))
            # find good coordinates
            if x + incX >= 3000:
                x = 20
                y += incY
                incY = int(sy)
            oglClass.SetPosition(x, y)
            x += incX
            self._pluginAdapter.addShape(shape=oglClass)
        self._pluginAdapter.refreshFrame()

    def _layoutLinks(self, oglLinks: OglLinks):

        # umlDiagram = umlFrame.GetDiagram()

        for oglLink in oglLinks:
            self._pluginAdapter.addShape(oglLink)

            # umlDiagram.AddShape(oglLink.sourceAnchor)
            # umlDiagram.AddShape(oglLink.destinationAnchor)

        self._pluginAdapter.refreshFrame()

    def __composeWildCardSpecification(self) -> str:

        inputFormat: InputFormat = self.inputFormat

        # wildcard: str = inputFormat.name + " (*." + inputFormat.extension + ")|*." + inputFormat.description
        wildcard: str = (
            f'{inputFormat.formatName} '
            f' (*, {inputFormat.extension}) '
            f'|*.{inputFormat.extension}'
        )
        return wildcard




from typing import cast

from logging import Logger
from logging import getLogger

from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface
from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat

from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import PluginName

from pyutplugins.ExternalTypes import OglObjects


FORMAT_NAME:        FormatName       = FormatName('Ascii')
PLUGIN_EXTENSION:   PluginExtension  = PluginExtension('acl')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('Export OGL as ASCII')


class IOAscii(IOPluginInterface):
    """
    Write ASCII and can read ASCII
    This just the skeleton.  Not sure if I want to do this
    """

    def __init__(self, pluginAdapter: IPluginAdapter):
        """

        Args:
            pluginAdapter:   A class that implements IPluginAdapter
        """
        super().__init__(pluginAdapter=pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._name      = PluginName('ASCII Class Export')
        self._author    = 'Philippe Waelti & Humberto A. Sanchez I>'
        self._version   = '2.0'

        self._inputFormat  = cast(InputFormat, None)
        self._outputFormat = OutputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)

    def setImportOptions(self) -> bool:
        """
        We do not support import
        Returns:  False
        """
        return True

    def setExportOptions(self) -> bool:
        return True

    def read(self) -> bool:
        return False

    def write(self, oglObjects: OglObjects):
        pass




from typing import Callable
from typing import cast

from logging import Logger
from logging import getLogger

from wx import ICON_ERROR
from wx import OK

from wx import MessageDialog
from wx import NewIdRef
from wx import BeginBusyCursor
from wx import EndBusyCursor

from wx import Yield as wxYield

from hasiihelper.Singleton import Singleton

from pyutplugins.preferences.PluginPreferences import PluginPreferences

from pyutplugins.plugininterfaces.ToolPluginInterface import ToolPluginInterface
from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface

from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.plugintypes.PluginDataTypes import ToolsPluginMap
from pyutplugins.plugintypes.PluginDataTypes import InputPluginMap
from pyutplugins.plugintypes.PluginDataTypes import OutputPluginMap
from pyutplugins.plugintypes.PluginDataTypes import PluginList
from pyutplugins.plugintypes.PluginDataTypes import PluginIDMap

from pyutplugins.ioplugins.IODTD import IODTD
from pyutplugins.ioplugins.IOGML import IOGML
from pyutplugins.ioplugins.IOJava import IOJava
from pyutplugins.ioplugins.IOPdf import IOPdf
from pyutplugins.ioplugins.IOPython import IOPython
from pyutplugins.ioplugins.IOWxImage import IOWxImage
from pyutplugins.ioplugins.IOXml import IOXml
from pyutplugins.ioplugins.IOMermaid import IOMermaid

from pyutplugins.toolplugins.ToolArrangeLinks import ToolArrangeLinks
from pyutplugins.toolplugins.ToolAscii import ToolAscii
from pyutplugins.toolplugins.ToolOrthogonalLayoutV2 import ToolOrthogonalLayoutV2
from pyutplugins.toolplugins.ToolSugiyama import ToolSugiyama
from pyutplugins.toolplugins.ToolTransforms import ToolTransforms


TOOL_PLUGIN_NAME_PREFIX: str = 'Tool'
IO_PLUGIN_NAME_PREFIX:   str = 'IO'


class PluginManager(Singleton):
    """
    Is responsible for:

    * Identifying the plugin loader files
    * Creating tool and Input/Output Menu ID References
    * Providing the callbacks to invoke the appropriate methods on the
    appropriate pyutplugins to invoke there functionality.

    Plugin Loader files have the following format:

    ToolPlugin=packageName.PluginModule
    IOPlugin=packageName.PluginModule

    By convention prefix the plugin tool module name with the characters 'Tool'
    By convention prefix the plugin I/O module with the characters 'IO'

    """
    IO_PLUGINS:   PluginList = PluginList([IOMermaid, IODTD, IOGML, IOJava, IOPdf, IOPython, IOWxImage, IOXml])
    TOOL_PLUGINS: PluginList = PluginList([ToolArrangeLinks, ToolAscii, ToolOrthogonalLayoutV2, ToolSugiyama, ToolTransforms])

    # noinspection PyAttributeOutsideInit
    def init(self,  *args, **kwargs):
        """
        Expects a pluginAdapter parameter in kwargs

        Args:
            *args:
            **kwargs:
        """

        self.logger: Logger = getLogger(__name__)

        self._pluginPreferences: PluginPreferences = PluginPreferences()
        # These are lazily built
        self._toolPluginsMap:   ToolsPluginMap   = ToolsPluginMap()
        self._inputPluginsMap:  InputPluginMap   = InputPluginMap()
        self._outputPluginsMap: OutputPluginMap  = OutputPluginMap()

        self._inputPluginClasses:  PluginList = cast(PluginList, None)
        self._outputPluginClasses: PluginList = cast(PluginList, None)

        self._pluginAdapter: IPluginAdapter = kwargs['pluginAdapter']

    @property
    def inputPlugins(self) -> PluginList:
        """
        Get the input pyutplugins.

        Returns:  A list of classes (the pyutplugins classes).
        """
        if self._inputPluginClasses is None:
            # noinspection PyAttributeOutsideInit
            self._inputPluginClasses = PluginList([])
            for plugin in PluginManager.IO_PLUGINS:
                pluginClass = cast(type, plugin)
                classInstance = pluginClass(None)
                if classInstance.inputFormat is not None:
                    self._inputPluginClasses.append(plugin)
        return self._inputPluginClasses

    @property
    def outputPlugins(self) -> PluginList:
        """
        Get the output pyutplugins.

        Returns:  A list of classes (the pyutplugins classes).
        """
        if self._outputPluginClasses is None:
            # noinspection PyAttributeOutsideInit
            self._outputPluginClasses = PluginList([])
            for plugin in PluginManager.IO_PLUGINS:
                pluginClass = cast(type, plugin)
                classInstance = pluginClass(None)
                if classInstance.outputFormat is not None:
                    self._outputPluginClasses.append(plugin)

        return self._outputPluginClasses

    @property
    def toolPlugins(self) -> PluginList:
        """
        Get the tool pyutplugins.

        Returns:    A list of classes (the pyutplugins classes).
        """
        return PluginManager.TOOL_PLUGINS

    @property
    def toolPluginsMap(self) -> ToolsPluginMap:
        if len(self._toolPluginsMap.pluginIdMap) == 0:
            self._toolPluginsMap.pluginIdMap = self.__mapWxIdsToPlugins(PluginManager.TOOL_PLUGINS)
        return self._toolPluginsMap

    @property
    def inputPluginsMap(self) -> InputPluginMap:
        if len(self._inputPluginsMap.pluginIdMap) == 0:
            self._inputPluginsMap.pluginIdMap = self.__mapWxIdsToPlugins(self.inputPlugins)
        return self._inputPluginsMap

    @property
    def outputPluginsMap(self) -> OutputPluginMap:
        if len(self._outputPluginsMap.pluginIdMap) == 0:
            self._outputPluginsMap.pluginIdMap = self.__mapWxIdsToPlugins(self.outputPlugins)
        return self._outputPluginsMap

    def doToolAction(self, wxId: int):
        """
        Args:
            wxId:   The ID ref of the menu item
        """
        pluginMap: PluginIDMap = self.toolPluginsMap.pluginIdMap

        # TODO: Fix this later for mypy
        clazz: type = pluginMap[wxId]   # type: ignore
        # Create a plugin instance
        pluginInstance: ToolPluginInterface = clazz(pluginAdapter=self._pluginAdapter)

        # Do plugin functionality
        BeginBusyCursor()
        try:
            pluginInstance.executeTool()
            self.logger.debug(f"After tool plugin do action")
        except (ValueError, Exception) as e:
            self.logger.error(f'{e}')
        EndBusyCursor()

    def doImport(self, wxId: int):
        """
        Args:
            wxId:       The ID ref of the menu item
        """
        idMap:        PluginIDMap    = self.inputPluginsMap.pluginIdMap
        clazz:        type              = idMap[wxId]     # type: ignore
        plugInstance: IOPluginInterface = clazz(pluginAdapter=self._pluginAdapter)
        self._doIOAction(methodToCall=plugInstance.executeImport)

    def doExport(self, wxId: int):
        """
        Args:
            wxId:       The ID ref of the menu item
        """
        idMap:        PluginIDMap  = self.outputPluginsMap.pluginIdMap
        clazz:        type              = idMap[wxId]     # type: ignore
        plugInstance: IOPluginInterface = clazz(pluginAdapter=self._pluginAdapter)
        self._doIOAction(methodToCall=plugInstance.executeExport)

    def _doIOAction(self, methodToCall: Callable):
        """
        Args:
            methodToCall:
        """

        try:
            wxYield()
            methodToCall()
        except (ValueError, Exception) as e:
            self.logger.error(f'{e}')
            booBoo: MessageDialog = MessageDialog(parent=None,
                                                  message=f'An error occurred while executing the selected plugin - {e}',
                                                  caption='Error!', style=OK | ICON_ERROR)
            booBoo.ShowModal()

    def __mapWxIdsToPlugins(self, pluginList: PluginList) -> PluginIDMap:

        pluginMap: PluginIDMap = cast(PluginIDMap, {})

        nb: int = len(pluginList)

        for x in range(nb):
            wxId: int = NewIdRef()

            pluginMap[wxId] = pluginList[x]

        return pluginMap




from typing import Callable
from typing import Dict
from typing import List
from typing import NewType
from typing import Union
from typing import cast

from dataclasses import dataclass
from dataclasses import field

from os import path as osPath

from enum import Enum

from wx import ClientDC

from pyutmodel.PyutLink import PyutLink

from ogl.OglClass import OglClass
from ogl.OglLink import OglLink
from ogl.OglInterface2 import OglInterface2
from ogl.OglNote import OglNote
from ogl.OglText import OglText

from ogl.OglUseCase import OglUseCase
from ogl.OglActor import OglActor

from ogl.sd.OglSDInstance import OglSDInstance
from ogl.sd.OglSDMessage import OglSDMessage

HybridLinks = Union[OglLink, OglInterface2]

OglClasses  = NewType('OglClasses',  List[OglClass])
OglLinks    = NewType('OglLinks',    List[HybridLinks])
OglNotes    = NewType('OglNotes',    List[OglNote])
OglTexts    = NewType('OglTexts',    List[OglText])
OglActors   = NewType('OglActors',   List[OglActor])
OglUseCases = NewType('OglUseCases', List[OglUseCase])

OglSDInstances = NewType('OglSDInstances', Dict[int, OglSDInstance])
OglSDMessages  = NewType('OglSDMessages',  Dict[int, OglSDMessage])

OglObjectType = Union[OglClass, OglLink, OglNote, OglText, OglActor, OglUseCase, OglInterface2, OglSDMessage, OglSDInstance, OglUseCase, OglActor]

OglObjects = NewType('OglObjects',  List[OglObjectType])
PyutLinks  = NewType('PyutLinks',   List[PyutLink])

SelectedOglObjectsCallback = Callable[[OglObjects], None]        # Todo: Figure out appropriate type for callback


def createOglObjectsFactory() -> OglObjects:
    """
    Factory method to create  the OglClasses data structure;

    Returns:  A new data structure
    """
    return OglObjects([])


@dataclass
class FrameSize:
    """
    The strategy is to provide minimal information to the pyutplugins
    we do not want them to not abuse it.
    """
    width:  int = -1
    height: int = -1


@dataclass
class FrameInformation:
    """
    The document title is the name of the frame
    """
    frameActive:        bool       = False
    selectedOglObjects: OglObjects = field(default_factory=createOglObjectsFactory)
    diagramTitle:       str         = ''
    diagramType:        str        = ''
    frameSize:          FrameSize  = FrameSize()
    clientDC:           ClientDC   = cast(ClientDC, None)


FrameInformationCallback = Callable[[FrameInformation], None]
FrameSizeCallback        = Callable[[FrameSize], None]


def createPluginClassesFactory() -> OglClasses:
    """
    Factory method to create  the OglClasses data structure;

    Returns:  A new data structure
    """
    return OglClasses([])


def createPluginLinksFactory() -> OglLinks:
    """
    Factory method to create  the OglLinks data structure;

    Returns:  A new data structure
    """
    return OglLinks([])


def createPluginNotesFactory() -> OglNotes:
    return OglNotes([])


def createOglTextsFactory() -> OglTexts:
    return OglTexts([])


def createPluginActorsFactory() -> OglActors:
    return OglActors([])


def createPluginUseCasesFactory() -> OglUseCases:
    return OglUseCases([])


def createPluginSDInstances() -> OglSDInstances:
    return OglSDInstances({})


def createPluginSDMessages() -> OglSDMessages:
    return OglSDMessages({})


PluginDocumentTitle = NewType('PluginDocumentTitle', str)


class PluginDocumentType(Enum):

    CLASS_DIAGRAM    = 'CLASS_DIAGRAM'
    SEQUENCE_DIAGRAM = 'SEQUENCE_DIAGRAM'
    USECASE_DIAGRAM  = 'USECASE_DIAGRAM'
    UNKNOWN_DIAGRAM  = 'UNKNOWN_DIAGRAM'

    @classmethod
    def toEnum(cls, enumStr: str) -> 'PluginDocumentType':

        assert (enumStr is not None) and (enumStr != ''), 'I need a real string dude'
        # TODO use switch statement when we get to 3.10.x
        if enumStr == 'CLASS_DIAGRAM':
            retEnum: PluginDocumentType = PluginDocumentType.CLASS_DIAGRAM
        elif enumStr == 'SEQUENCE_DIAGRAM':
            retEnum = PluginDocumentType.SEQUENCE_DIAGRAM
        elif enumStr == 'USECASE_DIAGRAM':
            retEnum = PluginDocumentType.USECASE_DIAGRAM
        else:
            assert False, 'Unknown diagram type'

        return retEnum

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()


@dataclass
class PluginDocument:
    """
    This is a plugin's version of a document
    """
    documentType:    PluginDocumentType  = PluginDocumentType.UNKNOWN_DIAGRAM
    documentTitle:   PluginDocumentTitle = PluginDocumentTitle('')
    scrollPositionX: int = -1
    scrollPositionY: int = -1
    pixelsPerUnitX:  int = -1
    pixelsPerUnitY:  int = -1
    oglClasses:      OglClasses     = field(default_factory=createPluginClassesFactory)
    oglLinks:        OglLinks       = field(default_factory=createPluginLinksFactory)
    oglNotes:        OglNotes       = field(default_factory=createPluginNotesFactory)
    oglTexts:        OglTexts       = field(default_factory=createOglTextsFactory)
    oglActors:       OglActors      = field(default_factory=createPluginActorsFactory)
    oglUseCases:     OglUseCases    = field(default_factory=createPluginUseCasesFactory)
    oglSDInstances:  OglSDInstances = field(default_factory=createPluginSDInstances)
    oglSDMessages:   OglSDMessages  = field(default_factory=createPluginSDMessages)


PluginDocuments     = NewType('PluginDocuments', dict[PluginDocumentTitle, PluginDocument])


def createPluginDocumentsFactory() -> PluginDocuments:
    return PluginDocuments({})


@dataclass
class PluginProject:
    """
    This is a plugin's version of a project
    """
    fileName:        str             = cast(str, None)
    projectName:     str             = cast(str, None)
    version:         str             = cast(str, None)
    codePath:        str             = cast(str, None)
    pluginDocuments: PluginDocuments = field(default_factory=createPluginDocumentsFactory)

    @classmethod
    def toProjectName(cls, fqFilename):
        """
        Return just the file name portion of the fully qualified path

        Args:
            fqFilename:  file name to display

        Returns:
            A project name
        """
        regularFileName: str = osPath.split(fqFilename)[1]
        projectName:     str  = osPath.splitext(regularFileName)[0]

        return projectName


CurrentProjectCallback     = Callable[[PluginProject], None]




from typing import cast
from typing import NewType
from typing import Tuple
from typing import List

from pyutplugins.toolplugins.sugiyama.SugiyamaConstants import H_SPACE
from pyutplugins.toolplugins.sugiyama.SugiyamaLink import SugiyamaLink

SugiyamaVertexEdge = NewType("SugiyamaVertexEdge", Tuple["SugiyamaNode", SugiyamaLink])
SugiyamaVEs        = NewType("SugiyamaVEs", List[SugiyamaVertexEdge])


class SugiyamaNode:
    """
    Real or virtual Sugiyama node interface.

    This class is an interface, you should not have an object of type
    SugiyamaNode. You have to use RealSugiyamaNode or VirtualSugiyamaNode
    objects.
    """
    def __init__(self):
        """
        """
        # Current barycenter value of the link, this value can be computed
        # on indexes or x coordinate. For more information, see function
        # getBarycenter()
        self.__barycenter: int = cast(int, None)
        self.__index: int | None = None
        """
        Index position on the level
        """
        self.__level: int | None = None
        """
        Index of level
        """
        self.__leftNode: SugiyamaNode | None = None
        """
        Node direct on the left on the same level
        """
        self.__rightNode = None
        """
        Node direct on the right on the same level
        """
        # parents and sons
        # ================
        #
        # A child is derived from a parent. There is a hierarchical link,
        # Realization or Inheritance, from source to parent.
        # Each node can have parents and children.
        self.__parents: SugiyamaVEs = SugiyamaVEs([])
        """
        List of parents
        """
        self.__children: SugiyamaVEs = SugiyamaVEs([])
        """
        List of children : [(SugiyamaNode, SugiyamaLink), ...]
        """
        self.__links:    SugiyamaVEs = SugiyamaVEs([])
        """
        List of non-hierarchical links : [(SugiyamaNode, SugiyamaLink), ...]
        """

    def getSize(self) -> Tuple[int, int]:
        """
        Get the size of the node.

        This function should be overridden.

        Returns:    (int, int): tuple (width, height)
        """
        return -1, -1

    def setPosition(self, x, y):
        """
        Set position of node.

        This function has to be implemented.

        Args:
            x: x position in absolute coordinates
            y: y position in absolute coordinates
        """
        pass

    def getPosition(self) -> Tuple[int, int]:
        """
        Get position of node.

        This function has to be implemented.  Should be a 'pass'  But Pycharm complains
        at method self.__pushToRight()

        Returns:    (int, int): tuple (x, y) is absolute coordinates

        """
        return 0, 0

    def addParent(self, parent: "SugiyamaNode", link: SugiyamaLink):
        """
        Update parent list

        Args:
            parent:  The parent
            link:   Link between self and parent
        """
        sugiyamaData: SugiyamaVertexEdge = cast(SugiyamaVertexEdge, (parent, link))
        self.__parents.append(sugiyamaData)

    def getParents(self) -> SugiyamaVEs:
        """
        Return list of parents

        Returns:  list of parents

        """
        return self.__parents

    def addChild(self, child, link):
        """
        Add a child.

        Args:
            child:  The child
            link:   Its Link
        """
        sugiyamaData: SugiyamaVertexEdge = cast(SugiyamaVertexEdge, (child, link))
        self.__children.append(sugiyamaData)

    def getChildren(self) -> SugiyamaVEs:
        """
        Get list of children.

        Returns:  The list of children
        """
        return self.__children

    def addNonHierarchicalLink(self, node: "SugiyamaNode", link: SugiyamaLink):
        """
        Add a non-hierarchical link, ie not a parent or child relation.

        Args:
            node:   linked node
            link:   link between the two nodes
        """
        vertexEdge: SugiyamaVertexEdge = SugiyamaVertexEdge((node, link))
        self.__links.append(vertexEdge)

    def getNonHierarchicalLink(self) -> SugiyamaVEs:
        """
        Get non-hierarchical links, ie not parent or child relationship

        Returns: [(SugiyamaNode, SugiyamaLink), ...] : list of tuple

        """
        return self.__links

    def setLevel(self, level: int):
        """
        Set level index.

        Args:
            level:   level index
        """
        self.__level = level

    def getLevel(self) -> int | None:
        """
        Get level index.

        Returns:    level index
        """
        return self.__level

    def setIndex(self, index: int):
        """
        Set index of node in level.

        Args:
            index: index of node
        """
        self.__index = index

    def getIndex(self) -> int | None:
        """
        Get index of node.

        Returns:    index of node
        """
        return self.__index

    def setLeftNode(self, node: 'SugiyamaNode'):
        """
        Set the left neighbor node.

        Args:
            node:   left neighbor
        """
        self.__leftNode = node

    def getLeftNode(self) -> "SugiyamaNode":
        """
        TODO: Fix this on 3.11 when allowed to return self
        Get the left neighbor node.

        Returns:    left neighbor
        """
        return self.__leftNode  # type: ignore

    def setRightNode(self, node: "SugiyamaNode"):
        """
        Set the right neighbor node.

        This method must be used before balancing the graph.

        Args:
            node: right neighbor
        """
        self.__rightNode = node

    def getRightNode(self) -> "SugiyamaNode":
        """
        TODO: Fix this on 3.11 when allowed to return self
        Get the right neighbor node.

        Returns:    right neighbor
        """
        return self.__rightNode     # type: ignore

    def getXMax(self) -> int:
        """
        Get bigger value of x coordinate according to right neighbor position.

        If there is no right neighbor, return None.

        Returns:    max x coordinate
        """
        if self.__rightNode is None:
            return cast(int, None)
        else:
            xRightNode = self.__rightNode.getPosition()[0]
            widthSelfNode = self.getSize()[0]
            return xRightNode - widthSelfNode - H_SPACE

    def __getAverageIndex(self, nodeList) -> int:
        """
        Compute the average of indices position on all the given nodes.

        Args:
            nodeList:    [SugiyamaNode, ...] : list of nodes

        Returns:  None if nodeList is empty.  float or None : Average of indexes position
        """
        if len(nodeList) == 0:
            return cast(int, None)
        else:
            summation = 0
            for (node, link) in nodeList:
                summation += node.getIndex()
            return summation // len(nodeList)

    def fixAnchorPos(self):
        """
        Fix the positions of the anchor points.

        This function has to be overloaded.

        """
        pass

    def upBarycenterIndex(self):
        """
        Compute the up barycenter value with fathers indexes.

        For reading this value, use getBarycenter()

        """
        self.__barycenter = self.__getAverageIndex(self.__parents)

    def downBarycenterIndex(self):
        """
        Compute the down barycenter value with sons indexes.

        For reading this value, use getBarycenter()

        """
        self.__barycenter = self.__getAverageIndex(self.__children)

    def barycenterIndex(self):
        """
        Compute the average of parents down-barycenter and sons up-barycenter.

        Before calling this function, you have to call downBarycenterIndex
        on each parent and upBarycenterIndex on each child.

        For reading this value, use getBarycenter()

        @author Nicolas Dubois
        """
        nodeList = self.__parents + self.__children
        if len(nodeList) == 0:
            # ~ print self.__index, "none"
            self.__barycenter = cast(int, None)
        else:
            summation = 0
            for (node, link) in nodeList:
                summation += node.getBarycenter()
            # ~ print self.__index, float(sum) / len(nodeList)
            self.__barycenter = summation // len(nodeList)

    def __getAverageX(self, nodeList: SugiyamaVEs) -> int:
        """
        Compute the average of x coordinates on all the given nodes.

        Args:
            nodeList:   [(SugiyamaNode, SugiyamaLink), ...] nodeList : parents or children

        Returns:    float or None : average of x coordinates
                    None if nodeList is empty.
        """
        if len(nodeList) == 0:
            return cast(int, None)
        else:
            summation: int = 0
            for (node, link) in nodeList:
                summation += node.getPosition()[0] + node.getSize()[0] // 2
            return summation // len(nodeList) - self.getSize()[0] // 2

    def upBarycenterX(self):
        """
        Compute the up barycenter value with fathers x coord.

        For reading this value, use getBarycenter()
        """
        self.__barycenter = self.__getAverageX(self.__parents)

    def downBarycenterX(self):
        """
        Compute the down barycenter value with sons x coord.

        For reading this value, use getBarycenter()
        """
        self.__barycenter = self.__getAverageX(self.__children)

    def barycenterX(self):
        """
        Compute the average of up and down barycenter on x coordinates.

        For reading this value, use getBarycenter()
        """
        parentsAndChildren: SugiyamaVEs = cast(SugiyamaVEs, self.__parents + self.__children)
        self.__barycenter = self.__getAverageX(parentsAndChildren)

    def getBarycenter(self) -> int:
        """
        Return the pre-computed barycenter value of the node.

        The barycenter value is computed when you call one of the following
        function:
            - upBarycenterIndex()
            - downBarycenterIndex()
            - upBarycenterX()
            - downBarycenterX()
        If you want to update the value, you have to recall one of the
        computing functions.

        Example:
            If you want the down barycenter computed on indexes:
            First call function downBarycenterIndex()
            and then call function getBarycenter()

        Returns:   float: barycenter value

        """
        return self.__barycenter

    def __computeWantedXPos(self):
        """
        Compute average of parents and children x coordinates.

        Returns:    tuple (int, int) : (x average, number of parents and children)
        """
        # Create list of parents and children
        parentsAndChildren = self.getParents() + self.getChildren()
        nbParentsAndChildren = len(parentsAndChildren)

        # If there are no parents and children
        if nbParentsAndChildren == 0:
            return None, 0

        # Compute average of parents and children x coordinates
        summation = 0
        # For all his parents and children
        for (node, link) in parentsAndChildren:
            # Get horizontal center coordinate of the node
            xCenterNode = node.getPosition()[0] + node.getSize()[0] // 2
            summation += xCenterNode

        return summation // nbParentsAndChildren - self.getSize()[0] // 2, nbParentsAndChildren

    def balance(self):
        """
        Compute a new x coordinate for balancing the hierarchical graph.

        Evaluate the best x coordinate for the node. If the best coord is on
        the right of the node, try to push the nodes on the right and then fix
        new position closer to best x coordinate.

        """
        # Get coordinate of node
        (x, y) = self.getPosition()

        # Evaluate best x coordinate of node
        (wantedXPos, nbFathersAndSons) = self.__computeWantedXPos()

        # If there are no parents nor sons
        if not wantedXPos:
            # Don't move
            return 0

        # If best x coord is righter than current x coordinate
        if wantedXPos > x:

            # If there is a right neighbor
            if self.__rightNode is not None:
                # Check max x coord according to right neighbor
                xMax = self.getXMax()

                # If right node is to close, try to push it
                if wantedXPos > xMax:
                    self.__rightNode.__pushToRight(
                        (wantedXPos - xMax) * nbFathersAndSons,
                        nbFathersAndSons)

                # Fix new position
                self.setPosition(min(wantedXPos, self.getXMax()), y)
            else:
                # There is no right node, fix position
                self.setPosition(wantedXPos, y)

        # Return true if node has been moved
        return self.getPosition()[0] - x > 3

    def __pushToRight(self, xDeltaSum: int, nbLinks: int):
        """
        Called by the left neighbor, that function tries to push the node to
        the right for the balancing. If there is not enough space, try to
        push the next node.

        xDelta is the ideal delta on x coordinate for reaching the best balance.

        Args:
            xDeltaSum:  is xDelta multiplied by the number of parents and sons
                which are pushing to the right.

            nbLinks:   is the number of parents and sons who push from the left.
        """
        # Get current position
        (x, y) = self.getPosition()
        # Get wanted x coordinate
        (wantedXPos, nbFathersAndSons) = self.__computeWantedXPos()

        # If the node has a barycenter value (has parents or sons)
        if wantedXPos is not None:
            # Add current delta in average
            xDeltaSum += (wantedXPos - x) * nbFathersAndSons
            # Update number of fathers and sons who are pushing
            nbLinks += nbFathersAndSons
        xDelta = xDeltaSum // nbLinks

        # If the node has to be moved to the right
        if xDelta > 0:
            # If the node has a right neighbor
            if self.__rightNode is not None:
                xMax = self.getXMax()
                # If we need more place, try to push the right neighbor
                if xMax < x + xDelta:
                    self.__rightNode.__pushToRight(
                        (x + xDelta - xMax) * nbLinks, nbLinks)
                # Fix the new position
                self.setPosition(min(x + xDelta, self.getXMax()), y)
            else:
                # No right neighbor, fix the new position
                self.setPosition(x + xDelta, y)




from typing import cast

from logging import Logger
from logging import getLogger

from ogl.OglLink import OglLink

from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.plugininterfaces.ToolPluginInterface import ToolPluginInterface

from pyutplugins.plugintypes.PluginDataTypes import PluginName

from pyutplugins.ExternalTypes import OglObjects


class ToolArrangeLinks(ToolPluginInterface):

    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._name      = PluginName('Arrange Links')
        self._author    = 'Cedric DUTOIT <dutoitc@shimbawa.ch>'
        self._version   = '1.1'

        self._menuTitle = 'Arrange links'

    def setOptions(self) -> bool:
        """
        Prepare the import.
        This can be used to ask some questions to the user.

        Returns:
            If `False`, the import is cancelled.
        """
        return True

    def doAction(self):

        self._pluginAdapter.selectAllOglObjects()
        self._pluginAdapter.getSelectedOglObjects(callback=self._doAction)

    def _doAction(self, oglObjects: OglObjects):

        modified: bool = False
        for oglObject in oglObjects:
            if isinstance(oglObject, OglLink):
                oglLink: OglLink = cast(OglLink, oglObject)
                self.logger.info(f"Optimizing: {oglLink}")
                oglLink.optimizeLine()
                modified = True
            else:
                self.logger.debug(f"No line optimizing for: {oglObject}")

        if modified is True:
            self._pluginAdapter.refreshFrame()
            self._pluginAdapter.indicatePluginModifiedProject()




from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import FormatName

from pyutplugins.plugintypes.BaseFormat import BaseFormat


class OutputFormat(BaseFormat):
    """
    Syntactic sugar
    """
    def __init__(self, formatName: FormatName, extension: PluginExtension, description: PluginDescription):
        super().__init__(formatName=formatName, extension=extension, description=description)




from typing import cast

from wx import Yield as wxYield

from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat
from pyutplugins.plugintypes.PluginDataTypes import PluginName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.SingleFileRequestResponse import SingleFileRequestResponse

from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglLinks
from pyutplugins.ExternalTypes import OglObjects

from pyutplugins.IPluginAdapter import IPluginAdapter
from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface

from pyutplugins.ioplugins.dtd.DTDParser import DTDParser

FORMAT_NAME:        FormatName        = FormatName("DTD")
PLUGIN_EXTENSION:   PluginExtension   = PluginExtension('dtd')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('W3C DTD 1.0 file format')


class IODTD(IOPluginInterface):

    def __init__(self, pluginAdapter: IPluginAdapter):
        super().__init__(pluginAdapter)

        # from super class
        self._name    = PluginName('IoDTD')
        self._author  = "C.Dutoit <dutoitc@hotmail.com>"
        self._version = '1.0'
        self._inputFormat  = InputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)
        self._outputFormat = cast(OutputFormat, None)

        self._fileToImport: str = ''

    def setImportOptions(self) -> bool:
        """
        We do need to ask for the input file name

        Returns:  'True', we support import
        """
        response: SingleFileRequestResponse = self.askForFileToImport(startDirectory=None)
        if response.cancelled is True:
            return False
        else:
            self._fileToImport = response.fileName

        return True

    def setExportOptions(self) -> bool:
        return False

    def read(self) -> bool:
        """

        Returns:  True if import succeeded, False if error or cancelled
        """
        filename: str = self._fileToImport

        dtdParser: DTDParser = DTDParser()

        dtdParser.open(filename=filename)

        oglClasses: OglClasses = dtdParser.oglClasses
        for oglClass in oglClasses:
            self._pluginAdapter.addShape(oglClass)

        oglLinks: OglLinks = dtdParser.links
        for oglLink in oglLinks:
            self._pluginAdapter.addShape(oglLink)

        self._pluginAdapter.refreshFrame()
        wxYield()
        self._pluginAdapter.indicatePluginModifiedProject()

        return True

    def write(self, oglObjects: OglObjects):
        """

        Args:
            oglObjects:

        Returns:  False, write not supported

        """
        return False




from typing import Dict
from typing import Optional

from logging import Logger
from logging import getLogger

from sys import platform as sysPlatform

from os import getenv as osGetEnv

from configparser import ConfigParser

from hasiihelper.Singleton import Singleton

from pyutplugins.toolplugins.orthogonal.LayoutAreaSize import LayoutAreaSize

from pyutplugins.ioplugins.mermaid.MermaidDirection import MermaidDirection


PLUGIN_PREFS_NAME_VALUES = Dict[str, str]


class PluginPreferences(Singleton):

    PREFERENCES_FILENAME:   str = 'pyutplugins.ini'
    THE_GREAT_MAC_PLATFORM: str = 'darwin'

    PYUT_PLUGINS_PREFERENCES_SECTION: str = 'PyutPlugins'
    DEBUG_SECTION:                    str = 'Debug'

    ORTHOGONAL_LAYOUT_SIZE:   str = 'orthogonal_layout_size'
    WX_IMAGE_FILENAME:        str = 'wx_image_filename'
    PDF_EXPORT_FILENAME:      str = 'pdf_export_filename'
    SUGIYAMA_STEP_BY_STEP:    str = 'sugiyama_step_by_step'
    MERMAID_LAYOUT_DIRECTION: str = 'mermaid_layout_direction'

    PLUGIN_PREFERENCES: PLUGIN_PREFS_NAME_VALUES = {
        ORTHOGONAL_LAYOUT_SIZE:   LayoutAreaSize(1000, 1000).__str__(),
        WX_IMAGE_FILENAME:        'WxImageDump',
        PDF_EXPORT_FILENAME:      'PyutExport.pdf',
        SUGIYAMA_STEP_BY_STEP:    'False',
        MERMAID_LAYOUT_DIRECTION: MermaidDirection.RightToLeft.value
    }

    DEBUG_TEMP_FILE_LOCATION: str = 'debug_temp_file_location'

    DEBUG_PREFERENCES: PLUGIN_PREFS_NAME_VALUES = {
        DEBUG_TEMP_FILE_LOCATION: 'False'
    }

    # noinspection PyAttributeOutsideInit
    def init(self, *args, **kwargs):
        self.logger: Logger = getLogger(__name__)

        self._config: ConfigParser = ConfigParser()

        self._preferencesFileName: str = self._getPreferencesLocation()

        self._loadPreferences()

    @property
    def orthogonalLayoutSize(self) -> LayoutAreaSize:

        serializedDimensions: str = self._config.get(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.ORTHOGONAL_LAYOUT_SIZE)
        return LayoutAreaSize.deSerialize(serializedDimensions)

    @orthogonalLayoutSize.setter
    def orthogonalLayoutSize(self, newValue: LayoutAreaSize):
        self._config.set(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.ORTHOGONAL_LAYOUT_SIZE, newValue.__str__())
        self._saveConfig()

    @property
    def wxImageFileName(self) -> str:
        return self._config.get(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.WX_IMAGE_FILENAME)

    @wxImageFileName.setter
    def wxImageFileName(self, newValue: str):
        self._config.set(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.WX_IMAGE_FILENAME, newValue)
        self._saveConfig()

    @property
    def pdfExportFileName(self) -> str:
        return self._config.get(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.PDF_EXPORT_FILENAME)

    @pdfExportFileName.setter
    def pdfExportFileName(self, newValue: str):
        self._config.set(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.PDF_EXPORT_FILENAME, newValue)
        self._saveConfig()

    @property
    def sugiyamaStepByStep(self) -> bool:
        ans: bool = self._config.getboolean(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.SUGIYAMA_STEP_BY_STEP)
        return ans

    @sugiyamaStepByStep.setter
    def sugiyamaStepByStep(self, newValue: bool):
        self._config.set(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.SUGIYAMA_STEP_BY_STEP, str(newValue))
        self._saveConfig()

    @property
    def mermaidLayoutDirection(self) -> MermaidDirection:
        directionStr: str = self._config.get(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.MERMAID_LAYOUT_DIRECTION)
        enumDirection: MermaidDirection = MermaidDirection.toEnum(directionStr)
        return enumDirection

    @mermaidLayoutDirection.setter
    def mermaidLayoutDirection(self, direction: MermaidDirection):
        self._config.set(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, PluginPreferences.MERMAID_LAYOUT_DIRECTION, direction.value)
        self._saveConfig()

    @property
    def debugTempFileLocation(self) -> bool:
        ans: bool = self._config.getboolean(PluginPreferences.DEBUG_SECTION, PluginPreferences.DEBUG_TEMP_FILE_LOCATION)
        return ans

    @debugTempFileLocation.setter
    def debugTempFileLocation(self, newValue: bool):
        self._config.set(PluginPreferences.DEBUG_SECTION, PluginPreferences.DEBUG_TEMP_FILE_LOCATION, str(newValue))
        self._saveConfig()

    def _getPreferencesLocation(self) -> str:

        if sysPlatform == "linux2" or sysPlatform == "linux" or sysPlatform == PluginPreferences.THE_GREAT_MAC_PLATFORM:
            homeDir:  Optional[str] = osGetEnv('HOME')
            fullName: str           = f'{homeDir}/{PluginPreferences.PREFERENCES_FILENAME}'
            preferencesFileName: str = fullName
        else:
            preferencesFileName = PluginPreferences.PREFERENCES_FILENAME

        return preferencesFileName

    def _loadPreferences(self):

        self._ensurePreferenceFileExists()

        # Read data
        self._config.read(self._preferencesFileName)
        self._addMissingPreferences()
        self._saveConfig()

    def _ensurePreferenceFileExists(self):

        try:
            # noinspection PyUnusedLocal
            with open(self._preferencesFileName, "r") as f:
                pass
        except (ValueError, Exception):
            try:
                with open(self._preferencesFileName, "w") as fw:
                    fw.write("")
                    self.logger.warning(f'Preferences file re-created')
            except (ValueError, Exception) as e:
                self.logger.error(f"Error: {e}")
                return

    def _addMissingPreferences(self):

        try:
            if self._config.has_section(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION) is False:
                self._config.add_section(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION)
            for prefName in PluginPreferences.PLUGIN_PREFERENCES:
                if self._config.has_option(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, prefName) is False:
                    self._addMissingPluginPreference(prefName, PluginPreferences.PLUGIN_PREFERENCES[prefName])

            if self._config.has_section(PluginPreferences.DEBUG_SECTION) is False:
                self._config.add_section(PluginPreferences.DEBUG_SECTION)
            for prefName in PluginPreferences.DEBUG_PREFERENCES:
                if self._config.has_option(PluginPreferences.DEBUG_SECTION, prefName) is False:
                    self._addMissingDebugPreference(prefName, PluginPreferences.DEBUG_PREFERENCES[prefName])

        except (ValueError, Exception) as e:
            self.logger.error(f"Error: {e}")

    def _addMissingPluginPreference(self, preferenceName: str, value: str):
        self._addMissingPreference(PluginPreferences.PYUT_PLUGINS_PREFERENCES_SECTION, preferenceName, value)

    def _addMissingDebugPreference(self, preferenceName: str, value: str):
        self._addMissingPreference(PluginPreferences.DEBUG_SECTION, preferenceName, value)

    def _addMissingPreference(self, sectionName: str, preferenceName: str, value: str):
        self._config.set(sectionName, preferenceName, value)
        self._saveConfig()

    def _saveConfig(self):
        """
        Save data to the preferences file
        """
        with open(self._preferencesFileName, "w") as fd:
            self._config.write(fd)




from dataclasses import dataclass

from pyutplugins.plugintypes.BaseRequestResponse import BaseRequestResponse


@dataclass
class ImportDirectoryResponse(BaseRequestResponse):
    directoryName: str = ''




from typing import List

from dataclasses import dataclass
from dataclasses import field

from pyutplugins.plugintypes.BaseRequestResponse import BaseRequestResponse


@dataclass
class MultipleFileRequestResponse(BaseRequestResponse):

    fileList:      List[str] = field(default_factory=list)
    directoryName: str       = ''




from typing import Callable

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from wx import EVT_TEXT
from wx import ID_ANY

from wx import SpinCtrl
from wx import CommandEvent

from wx import NewIdRef as wxNewIdRef

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

SPINNER_WIDTH:  int = 60
SPINNER_HEIGHT: int = 35


@dataclass
class SpinnerValues:
    value0: int = 0
    value1: int = 0


class DualSpinnerControl(SizedStaticBox):
    """
    A component that pairs two spinners in a horizontal panel;  Callers set
    the title to display and optionally the minimum and maximum spinner values
    Ideal for subclassing to use as a way to get x,y coordinates or width, height
    sizes
    """

    DEFAULT_MIN_VALUE: int = 100
    DEFAULT_MAX_VALUE: int = 300

    dscLogger: Logger = getLogger(__name__)     # Used as base class; So needs unique logger

    def __init__(self, sizedPanel: SizedPanel, boxTitle: str,
                 valueChangedCallback: Callable,
                 minValue: int = DEFAULT_MIN_VALUE, maxValue: int = DEFAULT_MAX_VALUE,
                 setControlsSize: bool = True,
                 ):
        """

        Args:
            sizedPanel   The parent panel
            boxTitle:    The text to display as the static box title
            valueChangedCallback:  The method to call when the value changes;  The method should expect the
                                   first parameter to be an object of type SpinnerValues
            minValue:       The minimum value for the spinner values
            maxValue:       The maximum value for the spinner values
            setControlsSize:  Whether to specify the spinner size;  This is a hack
            because in some SizedPanels the spinners are appropriately sized and in others they
            are not
        """

        super().__init__(sizedPanel, ID_ANY, boxTitle)

        self.SetSizerType('horizontal')
        # noinspection PyUnresolvedReferences
        # self.SetSizerProps(expand=True, proportion=1, border=(('left','right', 'bottom'),5))
        self.SetSizerProps(expand=True, proportion=1)

        self._callback: Callable = valueChangedCallback

        self._wxSpinner0Id: int = wxNewIdRef()
        self._wxSpinner1Id: int = wxNewIdRef()

        if setControlsSize is True:
            self._spinner0: SpinCtrl = SpinCtrl(self, self._wxSpinner0Id, "", size=(SPINNER_WIDTH, SPINNER_HEIGHT))
            self._spinner1: SpinCtrl = SpinCtrl(self, self._wxSpinner1Id, "", size=(SPINNER_WIDTH, SPINNER_HEIGHT))
        else:
            self._spinner0 = SpinCtrl(self, self._wxSpinner0Id, "")
            self._spinner1 = SpinCtrl(self, self._wxSpinner1Id, "")

        self._spinner0.SetRange(minValue, maxValue)
        self._spinner1.SetRange(minValue, maxValue)

        self._spinnerValues: SpinnerValues = SpinnerValues(minValue, maxValue)
        #
        # Bind to the text control;  Then we can type in or spin
        self.Bind(EVT_TEXT, self._onSpinnerValueChanged, self._spinner0)
        self.Bind(EVT_TEXT, self._onSpinnerValueChanged, self._spinner1)

    def _setSpinnerValues(self, spinnerValues: SpinnerValues):
        """
        Write only;  The appropriate way to retrieve the values is via the change callback
        Args:
            spinnerValues:
        """
        self._spinnerValues = spinnerValues
        self._spinner0.SetValue(spinnerValues.value0)
        self._spinner1.SetValue(spinnerValues.value1)
        self.dscLogger.info(f'range: {self._spinner0.GetRange()} - {self._spinner0.GetValue()=} {self._spinner1.GetValue()=}')

    spinnerValues = property(fset=_setSpinnerValues, doc='Write only property to initialize spinner values')

    def enableControls(self, value: bool):
        """
        Enable or disable the spinner controls

        Args:
            value: `True` to enable, else `False`
        """
        if value is True:
            self._spinner0.Enable()
            self._spinner1.Enable()
        else:
            self._spinner0.Disable()
            self._spinner1.Disable()

    def _onSpinnerValueChanged(self, event: CommandEvent):

        eventId:  int = event.GetId()
        newValue: int = event.GetInt()

        if eventId == self._wxSpinner0Id:
            self._spinnerValues.value0  = newValue
        elif eventId == self._wxSpinner1Id:
            self._spinnerValues.value1  = newValue
        else:
            self.dscLogger.error(f'Unknown spinner event id: {eventId}')

        self._callback(self._spinnerValues)




from enum import Enum


class MermaidDirection(str, Enum):

    diagramValue: str
    userValue:    str

    def __new__(cls, title: str, diagramValue: str) -> 'MermaidDirection':
        obj = str.__new__(cls, title)
        obj._value_ = title

        obj.diagramValue = diagramValue

        return obj

    RightToLeft = ('Right To Left', 'direction RL')
    LeftToRight = ('Left To Right', 'direction LR')

    @classmethod
    def toEnum(cls, enumStr: str) -> 'MermaidDirection':

        assert (enumStr is not None) and (enumStr != ''), 'I need a real string dude'
        match enumStr:
            case MermaidDirection.RightToLeft.value:
                retEnum: MermaidDirection = MermaidDirection.RightToLeft
            case MermaidDirection.LeftToRight.value:
                retEnum = MermaidDirection.LeftToRight
            case _:
                print(f'Unknown enumeration string {enumStr}')
                retEnum = MermaidDirection.LeftToRight

        return retEnum





class ALayoutNode:
    """
    ALayoutNode: Interface between OglObject/PyutObject and ALayout algorithm.
    """
    def __init__(self, oglObject):
        """

        Args:
            oglObject: interfaced ogl object
        """
        self._oglObject = oglObject

    def getSize(self):
        """
        Return the class size.

        Returns: (int, int): tuple (width, height)
        """
        return self._oglObject.GetSize()

    def getPosition(self):
        """
        Get class position.

        Returns: (int, int): tuple (x, y) coordinates
        """
        return self._oglObject.GetPosition()

    def setPosition(self, x: int, y: int):
        """
        Set the class position.

        Args:
            x: absolute coordinates
            y: absolute coordinates
        """
        self._oglObject.SetPosition(x, y)

    def getName(self) -> str:
        """
        Get the name of the class.

        Returns: name of the class
        """
        return self._oglObject.pyutObject.name




# Constants used in automatic layout

LEFT_MARGIN: int = 10    # Left margin
UP_MARGIN:   int = 10    # Up margin
H_SPACE:     int = 20     # Horizontal space (minimum value between nodes)
V_SPACE:     int = 50     # Vertical space (minimum value between nodes)




from typing import Dict
from typing import List
from typing import NewType
from typing import Tuple
from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from antlr4 import ParserRuleContext


from pyutplugins.ioplugins.python.pyantlrparser.Python3Parser import Python3Parser
from pyutplugins.ioplugins.python.pyantlrparser.Python3ParserVisitor import Python3ParserVisitor

MethodName          = NewType('MethodName', str)
PropertyName        = NewType('PropertyName', str)
ClassName           = NewType('ClassName', str)
ParentName          = NewType('ParentName', str)
ChildName           = NewType('ChildName', str)
MultiParameterNames = NewType('MultiParameterNames', str)  # comma separated parameter names
Field               = NewType('Field', str)
ExpressionText      = NewType('ExpressionText', str)
DataClassProperty   = NewType('DataClassProperty', Tuple[ClassName, ExpressionText])

ClassNames     = NewType('ClassNames', List[ClassName])
MethodNames    = NewType('MethodNames', List[MethodName])
ParameterNames = List[MultiParameterNames]
Fields         = List[Field]
Children       = List[Union[ClassName, ChildName]]

Methods    = NewType('Methods', Dict[ClassName, MethodNames])
Parameters = NewType('Parameters', Dict[MethodName, MultiParameterNames])
MethodCode = NewType('MethodCode', Dict[MethodName, List[str]])
Parents    = NewType('Parents', Dict[ParentName, Children])

PropertyNames      = Dict[PropertyName, ClassName]
PropertyParameters = Dict[PropertyName, ParameterNames]

DataClassProperties = List[DataClassProperty]


class PyutPythonVisitor(Python3ParserVisitor):

    PYTHON_SELF:         str = 'self'
    PYTHON_SELF_COMMA:   str = f'{PYTHON_SELF},'
    FIELD_IDENTIFIER:    str = f'{PYTHON_SELF}.'
    PYTHON_CONSTRUCTOR:  str = '__init__'
    PROPERTY_DECORATOR:  str = '@property'
    DATACLASS_DECORATOR: str = '@dataclass'

    NON_PROPERTY_INDICATOR: str = '\n'

    PYTHON_EQUALITY_TOKEN: str = '='

    def __init__(self):

        self.logger: Logger = getLogger(__name__)

        self.classNames:   ClassNames = ClassNames([])
        self.classMethods: Methods    = Methods({})
        self.parameters:   Parameters = Parameters({})
        self.methodCode:   MethodCode = MethodCode({})
        self.fields:       Fields     = []
        self._parents:     Parents    = Parents({})

        self.propertyNames:    PropertyNames = {}
        self.setterProperties: Parameters    = Parameters({})
        self.getterProperties: Parameters    = Parameters({})

        self.dataClassNames:      ClassNames          = ClassNames([])
        self.dataClassProperties: DataClassProperties = []

    @property
    def parents(self) -> Parents:
        return self._parents

    @parents.setter
    def parents(self, additionalParents: Parents):
        self._parents = additionalParents

    def visitFuncdef(self, ctx: Python3Parser.FuncdefContext):

        className = self._checkIfMethodBelongsToClass(ctx, Python3Parser.ClassdefContext)
        if className:
            methodName = ctx.getChild(1).getText()
            if not self.__isProperty(methodName):
                self.logger.debug(f'visitFuncdef: {methodName=}')
                if className not in self.classMethods:
                    self.classMethods[className] = MethodNames([methodName])
                else:
                    self.classMethods[className].append(methodName)

                self.__getMethodCode(methodName, ctx)

        return super().visitChildren(ctx)

    def visitDecorated(self, ctx: Python3Parser.DecoratedContext):

        funcDef  = ctx.funcdef()
        decorator: str = ctx.getChild(0).getText()
        #
        # Take care of data classes
        #
        if funcDef is not None:
            propName:  PropertyName = funcDef.getChild(1).getText()
            propCode:  str = ctx.getChild(1).getText()
            if self.__isPropertyDecorator(nameToCheck=decorator) is True:
                self.logger.info(f'visitDecorated - {decorator=} {propName=} {propCode=}')

                className: ClassName = self._checkIfMethodBelongsToClass(ctx, Python3Parser.ClassdefContext)

                self.logger.info(f'Update property names - {propName}')
                self.propertyNames[propName] = className
        elif self.__isDataClassDecorator(nameToCheck=decorator):
            className = ctx.classdef().getChild(1).getText()
            self.logger.info(f'visitDecorated -- We found dataclass: {className}')
            self.dataClassNames.append(className)

        return self.visitChildren(ctx)

    def visitClassdef(self, ctx: Python3Parser.ClassdefContext):

        className: ClassName = ClassName(ctx.getChild(1).getText())
        self.classNames.append(className)
        self.logger.debug(f'visitClassdef: Visited class: {className}')

        argListCtx: Python3Parser.ArglistContext = self._findArgListContext(ctx)
        if argListCtx is not None:
            self._createParentChildEntry(argListCtx, className)

        return super().visitClassdef(ctx)

    def visitParameters(self, ctx: Python3Parser.ParametersContext):

        if len(ctx.children) > 1:
            parameterNames: MultiParameterNames = ctx.getChild(1).getText()
            methodName:     MethodName          = self._getParametersMethodName(ctx.parentCtx)
            self.logger.debug(f'visitParameters: method {methodName=} {parameterNames=} ')

            if self.__isProperty(methodName=methodName):
                if parameterNames == PyutPythonVisitor.PYTHON_SELF:
                    self.getterProperties[methodName] = MultiParameterNames('')
                else:
                    strippedParameterNames: MultiParameterNames = MultiParameterNames(parameterNames.replace(PyutPythonVisitor.PYTHON_SELF_COMMA, ""))
                    self.setterProperties[methodName] = strippedParameterNames
            else:
                if parameterNames != PyutPythonVisitor.PYTHON_SELF:
                    strippedParameterNames = MultiParameterNames(parameterNames.replace(PyutPythonVisitor.PYTHON_SELF_COMMA, ""))
                    self.parameters[methodName] = strippedParameterNames

        return super().visitChildren(ctx)

    def visitExpr_stmt(self, ctx: Python3Parser.Expr_stmtContext):

        exprText: ExpressionText = cast(ExpressionText, ctx.getText())

        if exprText.startswith(PyutPythonVisitor.FIELD_IDENTIFIER) is True:
            areWeInInitMethod: bool = self.__isThisInitMethod(ctx)
            areWeAnAssignment: bool = self.__isThisAFieldAssignment(exprText)
            if areWeInInitMethod is True and areWeAnAssignment is True:
                self.logger.debug(f'Field expression: {exprText}')
                self.fields.append(Field(exprText.replace(PyutPythonVisitor.FIELD_IDENTIFIER, '')))
        else:
            isDataClass, className = self.__isThisADataClassProperty(ctx)
            if isDataClass is True:
                self.logger.info(f'Non-init:  {exprText=}')
                if PyutPythonVisitor.NON_PROPERTY_INDICATOR in exprText:
                    pass
                else:
                    dataClassName:     ClassName         = cast(ClassName, className)
                    dataClassProperty: DataClassProperty = DataClassProperty((dataClassName, exprText))
                    self.dataClassProperties.append(dataClassProperty)

        return super().visitChildren(ctx)

    def _checkIfMethodBelongsToClass(self, node: ParserRuleContext, classType) -> ClassName:

        while node.parentCtx:
            if isinstance(node, classType):
                return node.getChild(1).getText()
            node = node.parentCtx
        return cast(ClassName, None)

    def _getParametersMethodName(self, parentCtx: Python3Parser.FuncdefContext) -> MethodName:

        methodName: MethodName = parentCtx.getChild(1).getText()
        return methodName

    def _createParentChildEntry(self, parentCtx: Python3Parser.ArglistContext, childName: Union[ClassName, ChildName]):

        parentName: ParentName = cast(ParentName, parentCtx.getText())
        self.logger.debug(f'Class: {childName} is subclass of {parentName}')

        if parentName in self._parents:
            children: Children = self._parents[parentName]
            children.append(childName)
        else:
            children = [childName]

        self._parents[parentName] = children

    def __getMethodCode(self, methodName: MethodName, ctx: Python3Parser.FuncdefContext):

        methodText:     str       = ctx.getText()
        splitText:      List[str] = methodText.split('\n')
        justMethodCode: List[str] = splitText[1:len(splitText)]

        self.logger.debug(f'justMethodCode: {justMethodCode}')

        self.methodCode[methodName] = justMethodCode

    def _findArgListContext(self, ctx: Python3Parser.ClassdefContext) -> Python3Parser.ArglistContext:

        argListCtx: Python3Parser.ArglistContext = cast(Python3Parser.ArglistContext, None)
        for childCtx in ctx.children:
            if isinstance(childCtx, Python3Parser.ArglistContext):
                argListCtx = childCtx
                break

        return argListCtx

    def __isThisInitMethod(self, ctx: Python3Parser.Expr_stmtContext) -> bool:

        ans: bool = False

        methodCtx: Python3Parser.FuncdefContext = self.__findMethodContext(ctx)
        if methodCtx is not None:
            methodName: str = methodCtx.getChild(1).getText()
            if methodName == PyutPythonVisitor.PYTHON_CONSTRUCTOR:
                ans = True

        return ans

    def __isThisAFieldAssignment(self, exprText: str):

        ans: bool = False
        if PyutPythonVisitor.PYTHON_EQUALITY_TOKEN in exprText:
            ans = True

        return ans

    def __findMethodContext(self, ctx: Python3Parser.Expr_stmtContext) -> Python3Parser.FuncdefContext:

        currentCtx: ParserRuleContext = ctx

        while isinstance(currentCtx, Python3Parser.FuncdefContext) is False:
            currentCtx = currentCtx.parentCtx
            if currentCtx is None:
                break

        if currentCtx is not None:
            self.logger.debug(f'Found method: {currentCtx.getChild(1).getText()}')

        return cast(Python3Parser.FuncdefContext, currentCtx)

    def __isProperty(self, methodName: MethodName) -> bool:
        """
        Used by the function definition visitor to determine if the method name that is being
        visited has been marked as a property.

        Args:
            methodName:  The method name to check

        Returns: True if its is in our know list of property names
        """
        ans: bool = False

        if methodName in self.propertyNames:
            ans = True

        return ans

    def __isPropertyDecorator(self, nameToCheck: str) -> bool:

        ans: bool = False

        if nameToCheck.startswith(PyutPythonVisitor.PROPERTY_DECORATOR):
            ans = True
        else:
            strippedText: str = nameToCheck.lstrip('@')
            splitText: List[str] = strippedText.split('.')
            propName: str = splitText[0]
            if propName in self.propertyNames:
                ans = True

        return ans

    def __isDataClassDecorator(self, nameToCheck: str) -> bool:

        ans: bool = False

        if nameToCheck.startswith(PyutPythonVisitor.DATACLASS_DECORATOR):
            ans = True

        return ans

    def __isThisADataClassProperty(self, startCtx: ParserRuleContext) -> Tuple[bool, str]:
        """

        Args:
            startCtx:

        Returns:
            A Tuple that a boolean answering the question.  If the answer is true
            then the 2nd value of the tuple is the data class name
        """

        dataClassName: str = self.__getPotentialDataClassName(startCtx)

        ans: bool = False
        if dataClassName in self.dataClassNames:
            ans = True

        self.logger.info(f'isDataClass: {ans} - {dataClassName=}')

        return ans, dataClassName

    def __getPotentialDataClassName(self, startCtx: ParserRuleContext):

        potentialDataClassName: str               = ''
        parentCtx:              ParserRuleContext = startCtx

        while parentCtx is not None:
            # self.logger.info(f'{parentCtx=}')
            if isinstance(parentCtx, Python3Parser.ClassdefContext):
                potentialDataClassName = parentCtx.getChild(1).getText()
                self.logger.info(f'{potentialDataClassName=}')

            if isinstance(parentCtx, Python3Parser.DecoratedContext):
                decorator: str = parentCtx.getChild(0).getText()
                self.logger.info(f'visitSimple_stmt - {decorator=}  {potentialDataClassName=}')
            parentCtx = parentCtx.parentCtx

        return potentialDataClassName




from typing import cast

from pathlib import Path

from logging import Logger
from logging import getLogger

from wx import ICON_ERROR
from wx import MessageDialog
from wx import OK
from wx import Yield as wxYield

from pyutplugins.ExternalTypes import FrameInformation
from pyutplugins.ExternalTypes import OglObjects
from pyutplugins.IPluginAdapter import IPluginAdapter

from pyutplugins.ioplugins.mermaid.MermaidWriter import MermaidWriter

from pyutplugins.plugininterfaces.IOPluginInterface import IOPluginInterface
from pyutplugins.plugintypes.InputFormat import InputFormat
from pyutplugins.plugintypes.OutputFormat import OutputFormat
from pyutplugins.plugintypes.PluginDataTypes import FormatName
from pyutplugins.plugintypes.PluginDataTypes import PluginDescription
from pyutplugins.plugintypes.PluginDataTypes import PluginExtension
from pyutplugins.plugintypes.PluginDataTypes import PluginName
from pyutplugins.plugintypes.SingleFileRequestResponse import SingleFileRequestResponse


FORMAT_NAME:        FormatName        = FormatName('Mermaid Markdown')
PLUGIN_EXTENSION:   PluginExtension   = PluginExtension('md')
PLUGIN_DESCRIPTION: PluginDescription = PluginDescription('Export Ogl to Mermaid Markdown')


class IOMermaid(IOPluginInterface):

    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter=pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._requireSelection = False      # Override base class
        self._autoSelectAll    = True
        # from super class
        self._name    = PluginName('Mermaid Writer')
        self._author  = 'Humberto A. Sanchez II'
        self._version = '0.50'
        self._inputFormat  = cast(InputFormat, None)
        self._outputFormat = OutputFormat(formatName=FORMAT_NAME, extension=PLUGIN_EXTENSION, description=PLUGIN_DESCRIPTION)

        self._exportResponse: SingleFileRequestResponse = cast(SingleFileRequestResponse, None)
        self._oglObjects:     OglObjects                = cast(OglObjects, None)

    def setImportOptions(self) -> bool:
        return False

    def setExportOptions(self) -> bool:

        self._exportResponse = self.askForFileToExport()

        if self._exportResponse.cancelled is True:
            return False
        else:
            return True

    def read(self) -> bool:
        return False

    def write(self, oglObjects: OglObjects):

        self._oglObjects = oglObjects

        self._pluginAdapter.getFrameInformation(callback=self._frameInformationCallback)
        wxYield()

    def _frameInformationCallback(self, frameInformation: FrameInformation):

        if frameInformation.diagramType == 'CLASS_DIAGRAM':
            fqFileName: str = self._exportResponse.fileName
            self.logger.info(f'Export to {fqFileName=}')

            mermaidWriter: MermaidWriter = MermaidWriter(Path(fqFileName), writeCredits=True)
            mermaidWriter.translate(oglObjects=self._oglObjects)
        else:
            booBoo: MessageDialog = MessageDialog(parent=None, message='Currently Only Class Diagrams export is supported',
                                                  caption='Try Again!', style=OK | ICON_ERROR)
            booBoo.ShowModal()




from typing import cast

from logging import Logger
from logging import getLogger

from ogl.OglClass import OglClass
from ogl.OglInterface import OglInterface
from ogl.OglLink import OglLink
from ogl.OglLinkFactory import OglLinkFactory

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutLink import PyutLink
from pyutmodel.PyutLinkType import PyutLinkType


class LinkMakerMixin:
    """
    Used to aid the pyutplugins when they need to create a link;  Usually not used
    directly by the plugin but by the supplementary classes used by the pyutplugins

    """
    def __init__(self):

        self.logger: Logger = getLogger('LinkMakerMixin')

        self._oglLinkFactory: OglLinkFactory  = OglLinkFactory()

    def createLink(self, src: OglClass, dst: OglClass, linkType: PyutLinkType = PyutLinkType.INHERITANCE) -> OglLink:
        """
        Add a paternity link between child and father.

        Args:
            src:  subclass
            dst: Base Class
            linkType:   The type of link

        Returns: an OglLink

        """
        sourceClass:      PyutClass = cast(PyutClass, src.pyutObject)
        destinationClass: PyutClass = cast(PyutClass, dst.pyutObject)

        pyutLink: PyutLink = PyutLink("", linkType=linkType, source=sourceClass, destination=destinationClass)

        oglLink = self._oglLinkFactory.getOglLink(src, pyutLink, dst, linkType)

        src.addLink(oglLink)
        dst.addLink(oglLink)

        pyutClass: PyutClass = cast(PyutClass, src.pyutObject)
        pyutClass.addLink(pyutLink)

        return oglLink

    def createInterfaceLink(self, src: OglClass, dst: OglClass) -> OglInterface:
        """
        Adds an OglInterface link between src and dst.

        Args:
            src:    source of the link
            dst:    destination of the link

        Returns: the created OglInterface link
        """
        sourceClass:      PyutClass = cast(PyutClass, src.pyutObject)
        destinationClass: PyutClass = cast(PyutClass, dst.pyutObject)

        pyutLink:     PyutLink     = PyutLink(linkType=PyutLinkType.INTERFACE, source=sourceClass, destination=destinationClass)
        oglInterface: OglInterface = OglInterface(srcShape=src, pyutLink=pyutLink, dstShape=dst)

        src.addLink(oglInterface)
        dst.addLink(oglInterface)

        return oglInterface




from enum import Enum


class DTDElementTypes(Enum):

    EMPTY_ELEMENT         = 1
    ANY_CONTENT_ELEMENT   = 2
    PCDATA_ELEMENT        = 3
    MIN_SINGLE_OCCURRENCE = 4
    MIXED_CONTENT_ELEMENT = 5
    SEQUENCE_ELEMENT      = 6     # Sequence maybe a single element;  Zero or more is subtype 2

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'{self.name} - {self.__str__()}'




from typing import List
from typing import cast

from logging import Logger
from logging import getLogger

from os import O_CREAT
from os import O_WRONLY

from os import open
from os import write

from os import sep as osSep

from ogl.OglClass import OglClass

from pyutmodel.PyutLink import PyutLink
from pyutmodel.PyutLink import PyutLinks

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutLinkType import PyutLinkType
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutParameter import PyutParameter
from pyutmodel.PyutStereotype import PyutStereotype

from pyutplugins.ExternalTypes import OglClasses
from pyutplugins.ExternalTypes import OglObjects


class JavaWriter:

    def __init__(self, writeDirectory: str):
        """

        Args:
            writeDirectory:
        """
        self.logger:          Logger = getLogger(__name__)
        self._writeDirectory: str    = writeDirectory

        # defining Constants    TODO: Make REAL constants
        self.__tab = "    "
        self.__visibility = {
            "+": "public",
            "-": "private",
            "#": "protected"
        }

    def write(self, oglObjects: OglObjects):

        oglClasses: OglClasses = cast(OglClasses, [oglObject for oglObject in oglObjects if isinstance(oglObject, OglClass)])

        for oglClass in oglClasses:
            self.logger.debug(f'{oglClass=}')
            pyutClass: PyutClass = cast(PyutClass, oglClass.pyutObject)
            self._writeClass(pyutClass)

    def _writeClass(self, pyutClass: PyutClass):
        """
        Writing a class to a file.

        Args:
            pyutClass:  The PyutClass object to write
        """
        className = pyutClass.name

        # Opening a file for each class
        fqn:        str = f'{self._writeDirectory}{osSep}{className}.java'
        flags:      int = O_WRONLY | O_CREAT
        javaFileFD: int = open(fqn, flags)

        # Extract the data from the class
        fields  = pyutClass.fields
        methods = pyutClass.methods
        parents = pyutClass.parents

        allLinks: PyutLinks   = pyutClass.links

        stereotype: PyutStereotype = pyutClass.stereotype

        # List of links
        interfaces: PyutLinks = PyutLinks([])     # List of interfaces implemented by the class
        links:      PyutLinks = PyutLinks([])     # Aggregation and compositions

        self._separateLinks(allLinks, interfaces, links)

        # Is it an interface
        classInterface = "class"
        if stereotype.name is not None:
            stereotypeName: str = stereotype.name
            if stereotypeName.lower() == "interface":
                classInterface = "interface"

        self._writeClassComment(javaFileFD, className, classInterface)
        write(javaFileFD, f'public {classInterface} {className}'.encode())

        self._writeParent(javaFileFD, parents)
        self._writeInterfaces(javaFileFD, interfaces)
        write(javaFileFD, ' {\n\n'.encode())

        self._writeFields(javaFileFD, fields)

        # Aggregation and Composition
        self._writeLinks(javaFileFD, links)
        self._writeMethods(javaFileFD, methods)
        write(javaFileFD, '}\n'.encode())

    def _separateLinks(self, allLinks, interfaces, links):
        """
        Separate the different plugintypes of links into lists.

        Args:
            allLinks:   list of links of the class
            interfaces: list of interfaces implemented by the class
            links:

        Returns:

        """
        for link in allLinks:
            pyutLink: PyutLink = cast(PyutLink, link)
            linkType: PyutLinkType = pyutLink.linkType
            self.logger.debug(f'Found linkType: `{linkType}`')
            if linkType == PyutLinkType.INTERFACE:
                interfaces.append(link)
            elif linkType == PyutLinkType.COMPOSITION or linkType == PyutLinkType.AGGREGATION:
                links.append(link)

    def _writeClassComment(self, file: int, className, classInterface):
        """
        Write class comment with doxygen organization.

        Args:
            file:       file descriptor
            className:
            classInterface:
        """
        write(file, f'/**\n * {classInterface} {className}\n * Class information here \n */\n'.encode())

    def _writeFields(self, file: int, fields):
        """
        Write fields in file.

        Args:
            file:   file object to write to
            fields: list of all fields of a class
        """
        # Write fields header
        if len(fields) > 0:
            write(file, f'{self.__tab}// ------\n{self.__tab}// Fields\n{self.__tab}// ------\n\n'.encode())
        # Write all fields in file
        for field in fields:
            # Visibility converted from "+" to "public", ...
            visibility = self.__visibility[str(field.visibility)]

            # Type
            fieldType: str = str(field.type)    # TODO could just be field.type.value
            self.logger.debug(f'fieldType: {fieldType}')

            # Name
            name = field.name

            # Default value
            default = field.defaultValue
            if default is not None and default != "":
                if fieldType.lower() == 'string':
                    default = f' = "{default}"'
                else:
                    default = f' = {default}'
            else:
                default = ""

            # Comments
            if fieldType == "":
                comments = " // Warning: no type"
            else:
                comments = ""

            self._writeFieldComment(file, name, self.__tab)
            write(file, f'{self.__tab}{visibility} {fieldType} {name}{default};{comments}\n'.encode())

    def _writeFieldComment(self, file: int, name: str, tab=""):
        """
        Write method comment using doxygen format.

        Args:
            file:   File descriptor
            name:   The field name
            tab:    `tab` character to use
        """
        write(file, f'{tab}/**\n'.encode())
        write(file, f'{tab} * field {name}\n'.encode())
        write(file, f'{tab} * More field information here.\n'.encode())

        write(file, f'{tab} */\n'.encode())

    def _writeParent(self, file: int, parents):
        """
        Writing parent for inheritance.  (Java only has single inheritance)

        Args:
            file:       file descriptor
            parents:    list of parents
        """
        nbr = len(parents)

        # If there is a parent:
        if nbr != 0:
            write(file, " extends ".encode())

            # Only one parent allowed
            parent: PyutClass = parents[0]
            name:   str       = parent.name
            write(file, name.encode())

    def _writeInterfaces(self, file: int, interfaces: List[PyutLink]):
        """
        Writing interfaces implemented by the class.

        Args:
            file:       file descriptor
            interfaces: list of implemented interfaces
        """
        nbr = len(interfaces)

        # If there is at least one interface:
        if nbr != 0:
            write(file, " implements ".encode())

            # Write the first interface
            interfaceName: str = interfaces[0].getDestination().name
            write(file, interfaceName.encode())

            # For all next interfaces, write the name separated by a ','
            for interface in interfaces[1:]:
                write(file, f', {interface.getDestination().name}'.encode())

    def _writeLinks(self, file, links):
        """
        Write relation links in file.

        Args:
            file:   The file descriptor
            links:  The class links
        """
        write(file, "\n".encode())
        # Write all relation links in file
        for link in links:
            link = cast(PyutLink, link)
            # Get Class linked (type of variable)
            destinationLinkName = link.getDestination().name
            # Get name of aggregation
            name = link.name
            # Array or single variable
            if link.destinationCardinality.find('n') != -1 or link.destinationCardinality.find('*') != -1:
                array = "[]"
            else:
                array = ""

            write(file, f'{self.__tab}private {destinationLinkName} {name}{array};\n'.encode())

    def _writeMethods(self, file: int, methods):
        """
        Writing methods in source (.java) file

        Args:
            file:       file descriptor
            methods:    list of all method of a class
        """
        # Write header
        if len(methods) > 0:
            header: str = f'\n{self.__tab}// -------\n{self.__tab}// Methods\n{self.__tab}// -------\n\n'
            write(file, header.encode())

        # for all method in methods list
        for aMethod in methods:
            method: PyutMethod = cast(PyutMethod, aMethod)
            self.logger.debug(f'method: {method}')
            self._writeMethodComment(file, method, self.__tab)
            self._writeMethod(file, method)

    def _writeMethodComment(self, file: int, method: PyutMethod, tab=""):
        """
        Write method comment with doxygen organization.

        Args:
            file:   file descriptor
            method: pyutMethod
            tab:    tab character(s) to use

        """
        write(file, f'{tab}/**\n'.encode())
        write(file, f'{tab} * method {method.name}\n'.encode())
        write(file, f'{tab} * More info here.\n'.encode())

        for param in method.parameters:
            write(file, f'{tab} * @param {param.name} : {str(param.type)}\n'.encode())

        if str(method.returnType.value) != '':
            write(file, f'{tab} * @return {str(method.returnType)}\n'.encode())

        write(file, f'{tab} */\n'.encode())

    def _writeMethod(self, file: int, method: PyutMethod):
        """
        Writing a method in file : name(param, param, ...).

        Args:
            file:       File descriptor
            method:     method object
        """
        name:       str = method.name
        visibility: str = self.__visibility[str(method.getVisibility())]
        returnType: str = str(method.returnType)
        if returnType == "":
            returnType = "void"

        write(file, f'{self.__tab}{visibility} {returnType} {name}('.encode())

        # for all param
        nbParam = len(method.parameters)
        self.logger.debug(f'# params: {nbParam}')
        for param in method.parameters:
            # writing param
            self._writeParam(file, param)

            # comma between param
            nbParam = nbParam - 1
            if nbParam > 0:
                write(file, ' , '.encode())

        write(file, f') {{\n{self.__tab}}}\n\n'.encode())

    def _writeParam(self, file: int, param: PyutParameter):
        """
        Writing params to file.

        Args:
            file:   file descriptor
            param:  pyut parameter object to write
        """
        paramType: str = param.type.__str__()
        paramName: str = param.name
        write(file, f'{paramType} {paramName}'.encode())



# type: ignore
from typing import TextIO
from antlr4 import *
from antlr4.Token import CommonToken
from pyutplugins.ioplugins.python.pyantlrparser.Python3Parser import Python3Parser;

import sys
from typing import TextIO
import re


class Python3LexerBase(Lexer):
    NEW_LINE_PATTERN = re.compile('[^\r\n\f]+')
    SPACES_PATTERN = re.compile('[\r\n\f]+')

    def __init__(self, input: InputStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.tokens = []
        self.indents = []
        self.opened = 0
    
    def reset(self):
        self.tokens = []
        self.indents = []
        self.opened = 0
        super().reset()
    
    def emitToken(self, token):
        self._token = token
        self.tokens.append(token)
    
    def nextToken(self):
        # Check if the end-of-file is ahead and there are still some DEDENTS expected.
        if self._input.LA(1) == Python3Parser.EOF and len(self.indents) != 0:
            # Remove any trailing EOF tokens from our buffer.
            self.tokens = [token for token in self.tokens if token.type != Python3Parser.EOF]

            # First emit an extra line break that serves as the end of the statement.
            self.emitToken(self.commonToken(Python3Parser.NEWLINE, '\n'))

            # Now emit as much DEDENT tokens as needed.
            while len(self.indents) != 0:
                self.emitToken(self.createDedent())
                self.indents.pop()

            # Put the EOF back on the token stream.
            self.emitToken(self.commonToken(Python3Parser.EOF, '<EOF>'))

        next_ = super().nextToken()
        return next_ if len(self.tokens) == 0 else self.tokens.pop(0)

    def createDedent(self):
        return self.commonToken(Python3Parser.DEDENT, '')

    def commonToken(self, type_: int, text: str):
        stop = self.getCharIndex() - 1
        start = stop if text == '' else stop - len(text) + 1
        return CommonToken(self._tokenFactorySourcePair, type_, Lexer.DEFAULT_TOKEN_CHANNEL, start, stop)

    def getIndentationCount(self, whitespace: str):
        count = 0
        for c in whitespace:
            if c == '\t':
                count += 8 - count % 8
            else:
                count += 1
        return count

    def atStartOfInput(self):
        return self.getCharIndex() == 0

    def openBrace(self):
        self.opened += 1

    def closeBrace(self):
        self.opened -= 1

    def onNewLine(self):
        new_line = self.NEW_LINE_PATTERN.sub('', self.text)
        spaces = self.SPACES_PATTERN.sub('', self.text)

        # Strip newlines inside open clauses except if we are near EOF. We keep NEWLINEs near EOF to
        # satisfy the final newline needed by the single_put rule used by the REPL.
        next_ = self._input.LA(1)
        next_next = self._input.LA(2)

        if self.opened > 0 or (next_next != -1 and next_ in (10, 13, 35)):
            self.skip()
        else:
            self.emitToken(self.commonToken(Python3Parser.NEWLINE, new_line))
            indent = self.getIndentationCount(spaces)
            previous = 0 if len(self.indents) == 0 else self.indents[-1]

            if indent == previous:
                self.skip()
            elif indent > previous:
                self.indents.append(indent)
                self.emitToken(self.commonToken(Python3Parser.INDENT, spaces))
            else:
                while len(self.indents) > 0 and self.indents[-1] > indent:
                    self.emitToken(self.createDedent())
                    self.indents.pop()



from typing import cast
from typing import TextIO

from logging import Logger
from logging import getLogger

from math import floor
from math import ceil

from os import sep as osSep

from ogl.OglClass import OglClass

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutStereotype import PyutStereotype

from pyutplugins.plugininterfaces.ToolPluginInterface import ToolPluginInterface
from pyutplugins.plugininterfaces.IOPluginInterface import IPluginAdapter

from pyutplugins.plugintypes.ExportDirectoryResponse import ExportDirectoryResponse

from pyutplugins.plugintypes.PluginDataTypes import PluginName

from pyutplugins.ExternalTypes import OglObjects


class ToolAscii(ToolPluginInterface):
    """
    UML objects to an ASCII representation
    TODO:  This is really and IOPlugin that only does Output
    """

    def __init__(self, pluginAdapter: IPluginAdapter):

        super().__init__(pluginAdapter)

        self.logger: Logger = getLogger(__name__)

        self._name      = PluginName('ASCII Class Export')
        self._author    = 'Philippe Waelti <pwaelti@eivd.ch>'
        self._version   = '1.0'

        self._menuTitle = 'ASCII Class Export'

        self._exportDirectory: str = ''

    def setOptions(self) -> bool:

        response: ExportDirectoryResponse = self.askForExportDirectoryName(preferredDefaultPath=None)
        if response.cancelled is True:
            return False
        else:
            self._exportDirectory = response.directoryName
            self.logger.debug(f'selectedDir: {self._exportDirectory}')
            return True

    def doAction(self):
        self._pluginAdapter.getSelectedOglObjects(callback=self._doAction)

    def _doAction(self, selectedObjects: OglObjects):

        if len(selectedObjects) < 1:
            self.displayNoSelectedOglObjects()
            return
        self._write(selectedObjects)

    def _write(self, oglObjects: OglObjects):
        """
        Write the data to a file
        Args:
            oglObjects:   The objects to export
        """

        for oglObject in oglObjects:

            if not isinstance(oglObject, OglClass):
                continue

            pyutClass: PyutClass = cast(PyutClass, oglObject.pyutObject)
            filename:  str       = pyutClass.name

            fqFileName: str = f'{self._exportDirectory}{osSep}{filename}.acl'

            file: TextIO = open(f'{fqFileName}', "w")

            base = [pyutClass.name]
            pyutStereotype: PyutStereotype = pyutClass.getStereotype()
            if pyutStereotype is not None and pyutStereotype.name != '':
                base.append(str(pyutClass.getStereotype()))

            fields = [str(x) for x in pyutClass.fields]
            methods = [str(x) for x in pyutClass.methods]

            lineLength = max([len(x) for x in base + fields + methods]) + 4

            file.write(lineLength * "-" + "\n")

            for line in base:
                spaces = lineLength - 4 - len(line)
                file.write("| " + int(floor(spaces / 2.0)) * " " + line + int(ceil(spaces / 2.0)) * " " + " |\n")

            file.write("|" + (lineLength - 2) * "-" + "|\n")

            for line in fields:
                file.write("| " + line + (lineLength - len(line) - 4) * " " + " |\n")

            file.write("|" + (lineLength - 2) * "-" + "|\n")

            for line in methods:
                file.write("| " + line + (lineLength - len(line) - 4) * " " + " |\n")

            file.write(lineLength * "-" + "\n\n")

            file.write(pyutClass.description)

            file.close()




from typing import Callable
from typing import Union

from logging import Logger
from logging import getLogger

from ogl.OglDimensions import OglDimensions

from wx.lib.sized_controls import SizedPanel

from pyutplugins.common.ui.Dimensions import Dimensions
from pyutplugins.common.ui.DualSpinnerControl import SpinnerValues
from pyutplugins.common.ui.DualSpinnerControl import DualSpinnerControl


class DimensionsControl(DualSpinnerControl):
    """
    A facade around the basic dual spinner control;   Essentially
        * Converts the spinner values to and from the Dimension type;
        * Handles the spinner callback
        * Forwards the spinner values as Dimension values
    """
    DIMENSION_MIN_VALUE: int = 50
    DIMENSION_MAX_VALUE: int = 300

    def __init__(self, sizedPanel: SizedPanel, displayText: str,
                 valueChangedCallback: Callable,
                 minValue: int = DIMENSION_MIN_VALUE, maxValue: int = DIMENSION_MAX_VALUE,
                 setControlsSize: bool = True):
        """

        Args:
            sizedPanel             The parent window
            displayText:           The text to display as the position  title
            valueChangedCallback:  The method to call when the value changes;  The method should expect the
                                   first parameter to be a Dimension argument that is the new value
            minValue:              The minimum dimension value
            maxValue:              The maximum dimension value
            setControlsSize
        """
        self.logger:                     Logger     = getLogger(__name__)
        self._dimensionsChangedCallback: Callable   = valueChangedCallback
        self._dimensions:                Union[Dimensions, OglDimensions] = Dimensions()

        super().__init__(sizedPanel, displayText, self._onSpinValueChangedCallback, minValue, maxValue, setControlsSize)

    def _setDimensions(self, newValue: Union[Dimensions, OglDimensions]):
        self._dimensions  = newValue
        self.spinnerValues = SpinnerValues(value0=newValue.width, value1=newValue.height)

    dimensions = property(fset=_setDimensions, doc='Write only property to set dimensions on control')

    def _onSpinValueChangedCallback(self, spinnerValues: SpinnerValues):
        self.logger.info(f'{spinnerValues}')

        self._dimensions.width  = spinnerValues.value0
        self._dimensions.height = spinnerValues.value1

        self._dimensionsChangedCallback(self._dimensions)
